/* Copyright 2000 STMicroelectronics. Free for use under the Gnu Public License */

/* CAPTURE.C
   22/7/97
   capture module for cpia driver
*/
   
#include "warning.h"
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <toolhelp.h>
#include <mmsystem.h>

#pragma hdrstop

#include "vvl_defs.h"                    
#include "camtypes.h"
#include "cmd.h"
#include "pport.h"
#include "pplink.h"  
#include "cpialink.h"
#include "cpia_cmd.h"
#include "cpia_cam.h"

#include "debuglvl.h"
#include "tracemsg.h"

#include "capture.h"



#define OBS(x) MessageBox (NULL, "Function no longer used", x, MB_OK)
#define MB(x) MessageBox (NULL, x, "vvlcpia", MB_OK) 


static BOOL cmd_check( PCAPTURE pCap, LPSTR func );


/*-----------------------------------------------------------------*/

PCAPTURE capCreateCAPTURE( void )
{
 	HLOCAL hcap;
 	PCAPTURE pcap = NULL;
 
    hcap = LocalAlloc( LPTR, sizeof(CAPTURE) );
    
    if ( hcap )  
  	{
   		pcap = (PCAPTURE)LocalLock( hcap );
		pcap->HandleToThisInstance = hcap;
 		if ( pcap == NULL )
    	{
     		/* unable to allocate/lock the cpia struct */
			if ( hcap )
      		{
       			LocalUnlock( hcap );
       			LocalFree( hcap );
      		}                  
     		pcap = NULL;
    	}
	}
	
	if ( pcap ) 
	{
		/* initialise any member variables here */
	}

 	return pcap;
}


/*-----------------------------------------------------------------*/

BOOL capDestroyCAPTURE( PCAPTURE pCap )
{
	HLOCAL hcap;

	hcap = pCap->HandleToThisInstance;
	if ( hcap )
	{
		LocalUnlock( hcap );
		LocalFree( hcap );
	}                  

  	return TRUE;
}



/*-----------------------------------------------------------------*/
/* general initialisation. The camera has already been detected by this
   time. */
   
BOOL capInitialiseCAPTURE( PCAPTURE pCap, PLINK pLink )
{
 	BOOL res = FALSE;
 
	if ( pCap ) 
	{
	   	pCap->pLink = pLink;
	 	
   		pCap->CamError = FALSE;

	 	if ( GetCameraStatus( pCap->pLink ) )
	    {
	    	capSetDefaults( pCap );

		   	/* tweak the compression params slightly - only hysteresis is changed
      			from its default value */
   			CPIA_SetCompressionParams( pCap->pLink, 3, 9, 1, 3, 2, 5, 3, 2);  /* 0-13 */
 
			res = TRUE;
		} 
		else 
		{
	   		pCap->CamError = TRUE;
		}
	}    
	  
 	return res ;
}

/*-----------------------------------------------------------------*/    
BOOL capDeinitialiseCAPTURE( PCAPTURE pCap )
{
	/* nothing to do yet */ 
 	return( TRUE );
}

    
/*-----------------------------------------------------------------*/    


/*-----------------------------------------------------------------*/    


/*-----------------------------------------------------------------*/    
/* change the differencing thresholds on the camera. This does not affect
   the compression mode, and should only be called if the camera is in
   manual differencing mode
*/
   
BOOL capSetYUVThresholds( PCAPTURE pCap,
                              int YDiffThreshold, int UVDiffThreshold )
{
 	return capSetCompression( pCap, CAP_NULL, CAP_NULL,(WORD) YDiffThreshold,
                          (WORD) UVDiffThreshold );
}

/*-----------------------------------------------------------------*/    

BOOL capGetYUVThresholds( PCAPTURE pCap, int *YDiffThreshold, 
                              int *UVDiffThreshold )
{                             
 	BOOL res = FALSE;

 	return capGetCompression(pCap, NULL, NULL, 	(WORD *)YDiffThreshold, 
                   		       					(WORD *)UVDiffThreshold ); 
}


/*-----------------------------------------------------------------*/    



/*-----------------------------------------------------------------*/    
/* set the capture format on the camera.
   This doesn't alter the camera settings, which only happens when
   capSetFormat is called */

BOOL capSetSize( PCAPTURE pCap, BYTE CapSize )
{
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
   		pCap->CaptureSize = CapSize;
   		pCap->CaptureSettingsChanged = TRUE; 
   		pCap->chflags |= CPIA_SET_CAPTURE_SIZE | CPIA_SET_ROI;
   		res = TRUE;                          
  	}
  
 	return res;
}

/*-----------------------------------------------------------------*/    
/* return the capture format on the camera.
   either CAP_FMT_QCIF(0) or
          CAP_FMT_CIF(1)
*/
BYTE capGetSize( PCAPTURE pCap )
{
 	WORD res = 0;
 
 	if ( pCap )
  	{
   		res = pCap->CaptureSize;
  	}
  
 return res;
}                  

/*-----------------------------------------------------------------*/    
/* Set the subsample method and yuv order on the camera */

BOOL capSetSampleAndYUV( PCAPTURE pCap, BYTE subsample, BYTE yuvorder )
{
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
   		pCap->SubSample = subsample;
		pCap->YUVOrder = yuvorder;
   		pCap->CaptureSettingsChanged = TRUE;
   		pCap->chflags |= CPIA_SET_SAMPLE_YUV;
   		res = TRUE;                          
  	}
  
 	return res;
}

/*-----------------------------------------------------------------*/    
/* set the region of interest of the image.
   supplied parameters are in pixel dimensions. The sensor requires that
   x values are divided by 8, and y values by 4
   
   WORD xstart - start column in pixels
   WORD xend   - end column in pixels
   WORD ystart - start row in pixels
   WORD yend   - end row in pixels
*/
BOOL capSetROI( PCAPTURE pCap, WORD xstart, WORD xend, WORD ystart, 
                    WORD yend )
{
 	BOOL res = FALSE;
 
 	if (pCap && !pCap->CamError)
  	{
   		if ( pCap->ColStart !=(xstart/8) 
   			|| pCap->ColEnd !=(xend/8)
   			|| pCap->RowStart !=(ystart/4) 
   			|| pCap->RowEnd !=(yend/4) )
    	{
     		pCap->ColStart = xstart / 8;
     		pCap->ColEnd   = xend / 8;
     		pCap->RowStart = ystart / 4;
     		pCap->RowEnd   = yend / 4;
     		pCap->CaptureSettingsChanged = TRUE;
     		pCap->chflags |= CPIA_SET_ROI;
    	}
    
   		res = TRUE;
  	}
  
 	return res;
}


/*-----------------------------------------------------------------*/    
/* return the current RIO on the camera. The values returned are in
   pixels
*/
BOOL capGetROI( PCAPTURE pCap, WORD *xstart, WORD *xend, WORD *ystart, 
                    WORD *yend )
{
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
   		*xstart = pCap->ColStart * 8;
   		*xend   = pCap->ColEnd * 8;
   		*ystart = pCap->RowStart * 4;
   		*yend   = pCap->RowEnd * 4;
   		res = TRUE;                          
  	}
  
 	return res;
}


/*-----------------------------------------------------------------*/
/* set the compression mode on the camera.
   switching compression on requires using the local image buffer when 
   processing an image(this is dealt with in ppc.c)
*/
BOOL capSetCompression( PCAPTURE pCap, int compmode, int decimation, 
                        int ythresh, int uvthresh )
{             
 	BOOL res = FALSE;
                  
 	/* some parameter checking would be good here */
                  
 	if ( pCap )
  	{
   		if (compmode != CAP_NULL)
    		pCap->CompMode = (BYTE)compmode;
    
   		if (decimation != CAP_NULL)
    		pCap->Decimation = (BYTE)decimation;
    
   		if (ythresh != CAP_NULL)
    		pCap->YThresh = (BYTE)ythresh;
    
   		if (uvthresh != CAP_NULL)                                            
    		pCap->UVThresh = (BYTE)uvthresh;
    
   		pCap->CaptureSettingsChanged = TRUE;
   		pCap->chflags |= CPIA_SET_COMPRESSION;
   
 	  	res = TRUE;                          
  	}
  
 	return res;
}


/*-----------------------------------------------------------------*/
/* return the current compression details
*/
BOOL capGetCompression( PCAPTURE pCap, WORD *compmode, WORD *decimation,
                        WORD *ythresh, WORD *uvthresh )

{             
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
   		if (compmode)
    		*compmode = (WORD)pCap->CompMode;
    
   		if (decimation)
    		*decimation = (WORD)pCap->Decimation;
    
   		if (ythresh)
    		*ythresh = (WORD)pCap->YThresh;
    
   		if (uvthresh)
    		*uvthresh = (WORD)pCap->UVThresh;

 		res = TRUE;                          
  	}
  
 	return res;
}

/*-----------------------------------------------------------------*/
/* switch vertical decimation on or off */

BOOL capSetDecimation( PCAPTURE pCap, BOOL decim_on )
{
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
   		if (pCap->Decimation != decim_on)
    	{
     		pCap->Decimation = decim_on ? CAP_COMP_VDEC : CAP_COMP_NODEC;
     		pCap->CaptureSettingsChanged = TRUE;
     		pCap->chflags |= CPIA_SET_COMPRESSION;
    	}
   		res = TRUE;                          
  	}
  
 	return res;
}


/*-----------------------------------------------------------------*/
/* return TRUE if vertical decimation now on */

BOOL capGetDecimation( PCAPTURE pCap )
{
 	BOOL res = FALSE;
 
 	if (pCap)
  	{
   		res = pCap->Decimation == CAP_COMP_VDEC;
  	}
  
 	return res;
}




/* function below returns TRUE if it sent a command to the camera */
BOOL capDoFrameTimeProcessing( PCAPTURE pCap, BYTE *FrameHeader, 
								WORD AppFrameRequestTime, 
								WORD PreviousFrameProcessingTime ) 
{
	BOOL cmd_sent = FALSE;

	if ( !pCap->VidConf ) 
	{
		if ( PreviousFrameProcessingTime*3 < AppFrameRequestTime*2 )
		{ /* 
			we took less than 66% of the frame time to provide the image
			- therefore we can afford to send more data 
		  */
//			tracemsg( TRC_MSG, "Improve Quality\n" );
			ImproveQuality( pCap, TRUE );
		}
		else
		{
			ImproveQuality( pCap, FALSE );
		}
	}
	
	if ( pCap->CaptureSettingsChanged ) 
	{
		capSendCaptureParams( pCap );
		cmd_sent = TRUE;
	}

	return( cmd_sent );
}






/*-----------------------------------------------------------------*/    
/* set the capture settings for reading from the camera. Used to send 
   commands to the camera, preparing it for the upload.
   This might be a good place to send the ROI, format compression,
   ColourParams etc to the camera.
   Returns TRUE if it actually sent any commands to the camera.
*/

BOOL capSendCaptureParams( PCAPTURE pCap )
{
 	BOOL res = FALSE;
	BOOL sent_cmd = FALSE;
 	BYTE cmd;
	 
   	if ( pCap->CaptureSettingsChanged )
   	{ 
        /* one or more settings has changed. 
           Send this information to the camera
        */
        if (     pCap->chflags & CPIA_SET_FORMAT 
        	  || pCap->chflags & CPIA_SET_CAPTURE_SIZE
        	  || pCap->chflags & CPIA_SET_SAMPLE_YUV	)
        {
			sent_cmd = TRUE;
          	CPIA_SetFormat( pCap->pLink, pCap->CaptureSize, pCap->SubSample, pCap->YUVOrder );
          	if ( !cmd_check( pCap, "CPIA_setFormat") ) return FALSE;

			pCap->chflags &= ~(CPIA_SET_FORMAT | CPIA_SET_CAPTURE_SIZE | CPIA_SET_SAMPLE_YUV );
        }
        
        if ( pCap->chflags & CPIA_SET_ROI )
        {
			sent_cmd = TRUE;
          	CPIA_SetROI( pCap->pLink, pCap->ColStart, pCap->ColEnd, pCap->RowStart, pCap->RowEnd );
          	if ( !cmd_check( pCap, "capSetROI") ) return FALSE;

			pCap->chflags &= ~CPIA_SET_ROI;
        }

 		if (pCap->chflags & CPIA_SET_COMPRESSION)
        {
			sent_cmd = TRUE;
          	CPIA_SetCompression( pCap->pLink, pCap->CompMode, 
                               		pCap->Decimation );
          	if ( !cmd_check(pCap, "CPIA_SetCompression") ) return FALSE;
         
	        /* if running auto compression, set the auto values */
    		if (pCap->CompMode == CAP_COMP_AUTO)
           	{
            	/* any lingering Y and UV thresholds are used by the automatic
               	compression algorithm. Reset to default values to ensure a
               	decent picture */            
				sent_cmd = TRUE;
            	CPIA_SetYUVThresh( pCap->pLink, CAP_DFLT_YTHRESH, CAP_DFLT_UVTHRESH );
            	if ( !cmd_check(pCap, "CPIA_SetYUVThresh" ) ) return FALSE;
           	}
          
        	 /* if running manual compression, set the Y and UV thresholds */
         	if (pCap->CompMode == CAP_COMP_MANUAL)
           	{
				sent_cmd = TRUE;
            	CPIA_SetYUVThresh( pCap->pLink, pCap->YThresh, pCap->UVThresh );
            	if ( !cmd_check(pCap, "CPIA_SetYUVThresh") ) return FALSE;
        	}

			pCap->chflags &= ~CPIA_SET_COMPRESSION;
        }
 		
 		if (pCap->chflags & CPIA_SET_COMPRESSION_TARGET)
        {
			sent_cmd = TRUE;
           	CPIA_SetCompressionTarget( pCap->pLink, pCap->FRTargetting, 
                                       		pCap->TargetFR, pCap->TargetQ );
           	if ( !cmd_check(pCap, "CPIA_SetCompressionTarget") ) return FALSE;
        }

        pCap->CaptureSettingsChanged = FALSE;
	}
  
	/* if anything has gone wrong it will have bombed out before now ! */
 	return sent_cmd;
}


BOOL capDisableCompression( PCAPTURE pCap )
{
	BOOL res = FALSE;

 	res = CPIA_SetCompression( pCap->pLink, FALSE, FALSE );

    if ( pCap->chflags & CPIA_SET_FORMAT 
         || pCap->chflags & CPIA_SET_CAPTURE_SIZE
         || pCap->chflags & CPIA_SET_SAMPLE_YUV	)
    {
       	CPIA_SetFormat( pCap->pLink, pCap->CaptureSize, pCap->SubSample, pCap->YUVOrder );
       	if ( !cmd_check( pCap, "CPIA_setFormat") ) return FALSE;

		pCap->chflags &= ~(CPIA_SET_FORMAT | CPIA_SET_CAPTURE_SIZE | CPIA_SET_SAMPLE_YUV );
    }
        
    if ( pCap->chflags & CPIA_SET_ROI )
    {
       	CPIA_SetROI( pCap->pLink, pCap->ColStart, pCap->ColEnd, pCap->RowStart, pCap->RowEnd );
       	if ( !cmd_check( pCap, "capSetROI") ) return FALSE;

		pCap->chflags &= ~CPIA_SET_ROI;
    }

	pCap->CaptureSettingsChanged = TRUE;
	pCap->chflags |= CPIA_SET_COMPRESSION; /* force it to resend the compression parameters next time */

 	return( res );
}


/*-----------------------------------------------------------------*/    
/* initialise the CAPTURE struct with reasonable values */

void capSetDefaults( PCAPTURE pCap )
{
	pCap->CaptureSettingsChanged = TRUE;

	pCap->CaptureSize = CAP_FMT_CIF;   /* either CAP_FMT_CIF or CAP_FMT_QCIF */
	pCap->SubSample = CAP_FMT_YUV422;  /* 422 or 420 */
	pCap->YUVOrder = CAP_FMT_YUYV;
	                        
	/* CPIA_SetROI */                        
	pCap->ColStart = 0;                /* ROI parameters */
	pCap->ColEnd = 0; // CIF_COLS / 8;
	pCap->RowStart = 0;
	pCap->RowEnd = 0; // CIF_ROWS / 4;
	                     
	/* CPIA_SetCompression */
	pCap->CompMode = CAP_COMP_OFF;
	pCap->Decimation = CAP_COMP_NODEC;   /* CAP_COMP_VDEC */
	    
	/* CPIA_SetCompressionTarget */
	pCap->FRTargetting = CAP_TARGET_QUALITY;   /* quality */
	pCap->TargetFR = 15;                     /* target frame rate */
	pCap->TargetQ = 5;                       /* target quality */
	    
	/* CPIA_SetYUVThresh */
	pCap->YThresh = CAP_DFLT_YTHRESH;
	pCap->UVThresh = CAP_DFLT_UVTHRESH;
	    
	/* CPIA_SetTargetFPS */
	pCap->TargetFramerate = 30;		/* don't think this is actually still used */
}


/*-----------------------------------------------------------------*/    

BOOL GetCameraStatus(PLINK plink)

{
 BOOL res = TRUE;
 BYTE system_state,
      fatal_error;
 
/* return TRUE;*/
 
 CPIA_GetCameraStatus(plink,(BYTE FAR *) &system_state, NULL, NULL, 
                      (BYTE FAR *) &fatal_error, NULL, NULL,
                        NULL, NULL );

 if (system_state == UNINITIALISED_STATE /* || fatal_error != 0*/)
  res = FALSE;
  
 return res;
}


/*-----------------------------------------------------------------*/    
/* check that the previous camera command was accepted */
static BOOL cmd_check( PCAPTURE pCap, LPSTR func )
{
 	BYTE cmd = 0,
      	pwr = HI_POWER_STATE,
      	ftl = 0,
      	tmp;
 	BOOL res = TRUE;
 	char buf [255];

 	if ( pCap && pCap->CamError )
  		res = FALSE;
 	else
  	{                                      
   		/* check that the command took */    
   		tmp = CPIA_GetCameraStatus(pCap->pLink, &pwr, NULL, NULL, &ftl, 
                               &cmd, NULL, NULL, NULL );

   		if (pwr != HI_POWER_STATE || ftl != 0 || cmd != 0)
    	{
     		if ( pCap )
      			pCap->CamError = TRUE;
     		
	     	if (TMSET())
			{
				wsprintf( buf, "Error in %s, FatalError = 0x%02x, CmdError = 0x%02x%s",
	        		      		func, ftl, cmd, 
    		        	   		pwr != HI_POWER_STATE ?(LPSTR) " ** NOT IN HI POWER **" : (LPSTR) "" );

				tracemsg(TRC_MSG, buf );
			}

     		if ( cmd )
      			CPIA_ModifyCameraStatus(pCap->pLink, 4, ~cmd, 0 );

	    	res = FALSE;
    	}
    }
 	return res;
}


/*-----------------------------------------------------------------*/    
/* Set a compression target based on whether the driver is being used
   for video conferencing or not
   The values for targetting videoconferencing can be different from those
   when targetting framerate. vidconf determines which target is used
*/
BOOL capSetCompressionTarget( PCAPTURE pCap, BOOL vidconf, int confval,
                              int capval, int frate )
{                              
 	BOOL res = FALSE;
 
 	if ( pCap )
  	{
		if ( ((vidconf &&(confval == 10)) || (!vidconf &&(capval == 10)))
                && ((pCap->RowEnd - pCap->RowStart) > QCIF_ROWS/4) )
        {
            pCap->Decimation = TRUE;
        }
        else
        {
            pCap->Decimation = FALSE;
        }

		if ( (vidconf &&(confval == 1)) || (!vidconf &&(capval == 1)) ) 
		{
			/* disable all compression */
	    	 pCap->CompMode = CAP_COMP_OFF;
		}
		else
		{
			pCap->TargetFramerate= frate;
	   		pCap->VidConf = vidconf;
	   		SetCompressionTarget( pCap, vidconf, frate, vidconf ? confval : capval );
			pCap->chflags |= CPIA_SET_COMPRESSION_TARGET;
  		}


		pCap->CaptureSettingsChanged = TRUE;
		pCap->chflags |= CPIA_SET_COMPRESSION;
 	 
  		res = TRUE;                          
  	}
  
	return res;
}

/*-----------------------------------------------------------------*/    
/*
SetCompressionTarget:

An interface to CPIA_SetCompressionTarget for PPC2 and to the SetCompressionTarget command
for USB camera.

B.Stewart   4-9-97
Copyright 1997 VLSI Vision Limited.

Description - Control of compression is achieved via this function. It makes
a distinction between compression control for a video-conferencing application and
compression control for local display apps. It is the driver's responsibility to
determine which type of control to apply. As a clue however, we can assume that if
an app requests VfW streaming mode then it will be a video-conferencing app, otherwise
the app can be treated as some sort of local display application.

For video-conferencing the emphasis is to provide a fixed framerate which will be
requested by the video conferencing application. Call the function with confmode = TRUE
and the required framerate as frval. Quality is then under user control, use a range
of 1 to 10 for parameter qval. Within the function the values are mapped as follows:  

    val  1  2  3  4  5  6  7  8  9  10
    
    q    1  2  3  4  9  16 25 36 49 64

There is a linear portion(1:1 mapping) for low values follow by a squared relationship
to reach higher q values.

For local display the user is given full control of framerate vs.quality via one control.
Call the function with confmode = FALSE and a qval ranging from 1 to 10. This is mapped by
the function to the following range. The parameter frval is ignored in this case.
    
    val  1  2  3  4  5  6  7  8  9  10
    
    fr   1  4  9  12 15 18 21 24 27 30
    
So user control consists of one range control(i.e. slider) to control image quality in either
case of operation, video conf or local. In video-conf mode the driver is responsible for
determining desired frame, by consideration of the requests from the app it services. Indeed the
driver is also responsible for determining whether the mode should be video-conf or local.
*/


#define FIX_FR			1
#define FIX_Q			64
#define MAXRANGE		10
#define LINEAR_PORTION  4


static void SetCompressionTarget( PCAPTURE pCap, unsigned char confmode, 
                                  unsigned int frval, unsigned int qval )
{
static int fr,q=FIX_Q;
unsigned char target;

	if ( qval<1 || qval>10 )
		return;

	if ( !confmode )
	{	/* == local display == */

		target = CAP_TARGET_FRAMERATE;
		/* == ignore the framerate(frval) setting == */

		/* 
		although this is a quality range it is actually achieved
		by varying the target framerate
		*/

		/* == convert qval to a frammerate == */
		if ( qval == 1 )
			fr = 1;	/* == take care of extreme value == */
		else
			if (qval == 2)
				fr = 6;
			else
				fr = qval*4;

		if ( fr > 28 ) fr = 28;

		q = FIX_Q;	
	}
	else
	{	/* == conferencing == */
		
		/* == emphasis on providing fixed framerate == */
		/* == target quality or framerate == */
		target = CAP_TARGET_FRAMERATE;
		/* == map frval to framerate == */
		fr = frval;	
		
		if ( fr > 12 ) fr = 12;	  /* really aim for 15 fps maximum */

		/* == map qval to quality value == */
		if ( qval<LINEAR_PORTION )
			q = qval * 2;
		else
			q =(qval - LINEAR_PORTION/4)*(qval - LINEAR_PORTION/4);
	}

 	pCap->FRTargetting = target;
 	pCap->TargetFR = fr;
 	pCap->TargetQ = q;

	return;
}
												 



#define Q_STEP			8
#define Q_MAX			64
#define Q_MIN			4


void ImproveQuality( PCAPTURE pCap, BOOL Improve )
{
	BOOL changed = FALSE;

	if ( Improve ) 
	{
//		tracemsg( TRC_MSG, "ImproveQuality requested\n" );

	 	if ( pCap->TargetQ > Q_MIN ) 
	 	{
			/* trick bit to ensure Q can always reach MIN_Q */
			if ( pCap->TargetQ < (Q_MIN + Q_STEP) ) 
				pCap->TargetQ = Q_MIN;
			else
				pCap->TargetQ -= Q_STEP;

			changed = TRUE;
		}
	}
	else 
	{
		if ( pCap->TargetQ < Q_MAX )
		{
			pCap->TargetQ += Q_STEP * 4;
			changed = TRUE;
		}
	}

	if ( changed )
	{
		pCap->chflags |= CPIA_SET_COMPRESSION_TARGET;
		pCap->CaptureSettingsChanged = TRUE;
	}

	return;
}


