/* Copyright 2000 STMicroelectronics. Free for use under the Gnu Public License */
// DEVLOAD.CPP - utility routines for VxD Loader/Unloader app
// Copyright (c) 1995, Vireo Software, Inc.

// This module is designed for inclusion in a 16-bit medium model
// Windows app. It provides routines for calling VXDLDR to load
// and unload dynamically loadable VxDs.

// The VXDLDR is provided with Windows for Workgroups 3.11 and Windows 95

//#include "stdafx.h"
#include <windows.h>

#include "devload.h"

int (far*pVXDLDR)()=NULL;               // address of VXDLDR API entry point
WORD ourDx;

//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
// Function
//      MakeSureVXDLDRIsPresent - test presence of VXDLDR and initialize
//
// Remarks
//      Call this routine during initialization of the app. It will determine
//      if the VXDLDR device is present. If not, you must add the line
//      device=vxdldr.386 to the [386enh] section of system.ini.
//
// Returns
//      Returns TRUE if the VXDLDR is present; otherwise FALSE.
//
BOOL MakeSureVXDLDRIsPresent()
{
        if (!pVXDLDR)
        {
                pVXDLDR = LocateVxD(VXDLDR_DEVICE_ID);
                if (!pVXDLDR)
                        return FALSE;
        }
        return TRUE;
}

//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
// Function
//      LoadVxD - load a dynamically loadable VxD
//
// Parameters
//      vxdname         null terminated full pathname of a dynamically
//                      loadable VxD (e.g. "c:\\xyz\\pqr\\abc.386")
// Returns
//      Returns VXDLDR status code. A value of zero means success. For
//      other errors, see devload.h. The value can be passed to
//      GetVXDLDRStatusString (below) to get an ascii string explaining the
//      error.
//
int LoadVxD(const char* vxdname)
{
        if (!MakeSureVXDLDRIsPresent())
                return VXDLDR_ERR_NOT_LOADED;

		ourDx = OFFSETOF(vxdname);

        _asm mov dx, ourDx
        _asm mov ax, CMD_LOAD
        return pVXDLDR();
}

//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
// Function
//      UnloadVxD - unload a previously loaded dynamically loadable VxD
//
// Parameters
//      vxdname         Device name of the VxD to unload. Unlike LoadVxD,
//                      you pass the name of the device (e.g. "VPICD"), not 
//                      the name of of the file containing the VxD.
//                      
// Returns
//      Returns VXDLDR status code. A value of zero means success. For
//      other errors, see devload.h. The value can be passed to
//      GetVXDLDRStatusString (below) to get an ascii string explaining the
//      error.
//
int UnloadVxD(const char* vxdname)
{
        if (!MakeSureVXDLDRIsPresent())
                return VXDLDR_ERR_NOT_LOADED;
 	
 		ourDx = OFFSETOF(vxdname);

        _asm mov dx, ourDx
        _asm mov ax, CMD_UNLOAD
        _asm mov bx, 0
        return pVXDLDR();
}

//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
// Function
//      GetVXDLDRStatusString - get string for VXDLDR status code
//
// Parameters
//      status          status code returned from VXDLDR
//
// Returns
//      Returns a pointer to a null terminated string for the given status
//      code.
//
const char* GetVXDLDRStatusString(int status)
{
        const char* result;

        switch (status)
        {
        case VXDLDR_ERR_NOT_LOADED:
                result = "The VXDLDR is not present";
                break; 
        case 0:
                result = "Success";
                break;
        case VXDLDR_ERR_OUT_OF_MEMORY:
                result = "Out of memory";
                break;
        case VXDLDR_ERR_IN_DOS:
                result = "Error in DOS";
                break;
        case VXDLDR_ERR_FILE_OPEN_ERROR:
                result = "Error opening file";
                break;
        case VXDLDR_ERR_FILE_READ:
                result = "Error reading file";
                break;
        case VXDLDR_ERR_DUPLICATE_DEVICE:
                result ="Device already loaded";
                break;
        case VXDLDR_ERR_BAD_DEVICE_FILE:
                result = "Not a valid dynamically loadable VxD";
                break;
        case VXDLDR_ERR_DEVICE_REFUSED:
                result = "Device refused to load";
                break;
        case VXDLDR_ERR_NO_SUCH_DEVICE:
                result = "No such device";
                break;
        case VXDLDR_ERR_DEVICE_UNLOADABLE:
                result="Device is not unloadable";
                break;
        case VXDLDR_ERR_ALLOC_V86_AREA:
                result = "Error allocating V86 area";
                break;
        case VXDLDR_ERR_BAD_API_FUNCTION:
                result= "Internal error: bad API";
                break;
        default:
                result = "Unknown error";
                break;
        }

        return result;
}

//-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/
// Function
//      LocateVxD - find the API callback address for a VXD
//
// Parameters
//      devID           The 16-bit device ID for the VxD whose API entry
//                      is to be returned.
//
// Remarks
//      This function issues an INT 2fh with AX=1684h and BX=deviceID. If
//      the indicated device is present and has an API entry point for the
//      current mode, then a callback address is returned in ES:DI. 
//
// Returns
//      Returns the callback address, or NULL if the device is not found.
//
int (far *LocateVxD(WORD devID))()
{
        int (far* p)();

        _asm {
                xor di, di
                mov es, di
                mov ax, 1684h
                mov bx, devID
                int 2Fh
                mov word ptr [p], di
                mov word ptr [p+2], es
        }

        return p;
}
