/* Copyright 2000 STMicroelectronics. Free for use under the Gnu Public License */
//
//	File		dma_ctrl.h
//	Purpose		provides dma services to fill a buffer
//
//	Created by 	S. Gillies
//	Copyright	(c) 1997 VLSI Vision Ltd, all rights reserved.
//
//	Revision history	
//		0-01	16/12/97	SG	Created
//		0-02	24/12/97	SG	New DMA_SetMode() parameter
//		0-03	19-01-98	SG	A physical address aquired from a VxD can now be used
//								bufferType parameter added to DMA_Init
//								Physical parameter added to DMA_StartTransfer
//		0-06	10-02-98	SG	FreeBuffer added
//								SetBuffer parameter Size_1 added
//		
//		
#ifndef _DMA_CTRL_H_
#define _DMA_CTRL_H_

#ifdef __cplusplus
extern "C" {
#endif 


//
//	System Includes
//
#ifdef _WINDOWS
#include <windows.h>
#endif


//
//	Local Includes
//
#include "vvl_defs.h"


//
//	Definitions for use with public functions
//
#define DMA_MODE_DEMAND     0x00	//	Continual transfer while DREQ asserted and TC not reached
#define DMA_MODE_SINGLE     0x40	//	Host must pulse DREQ for each transfer
#define DMA_MODE_BLOCK      0x80	//	Continual transfer once DREQ has been detected until TC
#define DMA_MODE_CASCADE    0xc0 

#define DMA_MODE_ADDRINC    0x00
#define DMA_MODE_ADDRDEC    0x20

#define DMA_MODE_AUTOINIT   0x10
#define DMA_MODE_NOAUTOINIT 0x00

#define DMA_MODE_VERIFY		0x00
#define DMA_MODE_WRITE		0x04
#define DMA_MODE_READ		0x08

typedef enum {
	DMA_BUFFER_NONE,
	DMA_BUFFER_PM_APP,
	DMA_BUFFER_DOS_APP,
	DMA_BUFFER_VDS,
	DMA_BUFFER_PHYS
} T_DMA_BUFFER;

//
//	Public function declarations
//

BOOL DMA_Init( T_DMA_BUFFER BufferType );
//
//	Resets internal parameters to default settings
//		Default Mode = DMA_MODE_DEMAND | DMA_MODE_WRITE | DMA_MODE_NOAUTOINIT | DMA_MODE_ADDRINC    
//
//	Parameters
//		BufferType	type of buffer to dma transfer data into
//
//	Returns
//		TRUE if _WINDOWS and VDS installed, OR not _WINDOWS
//		FALSE if _WINDOWS and VDS is not installed
//

BOOL DMA_SetBuffer( BYTE Channel, LPBYTE Buffer, DWORD Physical, WORD Size_1 );
//
//	Sets a buffer to be used for dma transfers.
//
//	Parameters
//		Channel		dma channel to use
//		Buffer		pointer to dma buffer
//		Physical	physical address of dma buffer
//		Size_1		size of buffer in bytes - 1
//
//	Returns
//		TRUE if buffer data set up OK
//		FALSE if otherwise

void DMA_FreeBuffer( BYTE Channel );
//
//	Resets module dma buffer information
//
//	Parameters
//		Channel		dma channel to use
//	

BOOL DMA_StartTransfer( BYTE Channel, WORD Bytes_1 );
//
//	Programs the DMA controller to perform a data read.
//	The peripheral may also need to be configured for DMA transfer, e.g.
//	set a parallel port to ECP mode and enable ECP DMA.
//
//	Parameters
//		Channel		dma channel to use
//		Bytes_1		number of bytes to transfer - 1
//		Buffer		far pointer to buffer to fill when using bufferTypes:
//						DMA_BUFFER_PM_APP, DMA_BUFFER_DOS_APP, DMA_BUFFER_VDS
//		Physical	physical address of dma buffer when using buffer type:
//						DMA_BUFFER_PHYS
//	Returns
//		TRUE if transfer started OK
//		FALSE if transfer failed to start for any reason
//

void DMA_EndTransfer( BYTE Channel );
//
//	Ends a dma transfer, copying data to the user's buffer if necessary
//	This function is called when the DMA transfer is complete, which can
//	be confirmed by calling DMA_CheckTCBit()
//	The peripheral may also need to be configured out of DMA mode
//
//	Parameters
//		Channel		dma channel to use
//	Returns
//		TRUE if transfer is ended successfully and current buffer has been filled
//		FALSE otherwise
//

void DMA_AbortTransfer( BYTE Channel );
//
//	Disables the DMA controller
//	The peripheral may also need to be configured out of DMA mode
//
//	Parameters
//		Channel		dma channel to abort
//

BOOL DMA_CheckTCBit( BYTE Channel );
//
//	Checks to see if TC bit is set, indicating that the current DMA transfer is
//	complete
//
//	Parameters
//		Channel		dma channel to check
//	Returns
//		TRUE if TC bit is set
//		FALSE if TC bit is not set or channel is invalid
//

void DMA_SetMode( BYTE Mode );
//
//	Sets current transfer mode
//
//	Parameters
//		Mode	mode flags - bitwise OR (|) DMA_MODE_* macros defined above to
//				create a mode byte
//                                


#ifdef __cplusplus
}
#endif 

#endif 
