/* Copyright 2000 STMicroelectronics. Free for use under the Gnu Public License */
/*
 	Version 1-10
	
	pport.c

	Project: Vision DUAL Camera

	interface to parallel port, takes care of reading
	writing data and the operation mode of the port
	
*/

#include <string.h>

#include <conio.h>	/* needed for intrinsic declarations of _inp() / _outp() 		*/

#ifdef _WINDOWS
#include <windows.h>
#include <mmsystem.h>
#include "win_vect.h"
#else
#include <time.h>
#include <dos.h>   /* declarations of _dos_getvect() / _dos_setvect() / MK_FP()	*/
#define DosGetIntVector _dos_getvect
#define DosSetIntVector _dos_setvect
#define MAKELP  		MK_FP
#endif

 


#include "vvl_defs.h"
#include "dma_ctrl.h"

#include "pport.h"

/* disable all optimisation == /Od */


#pragma intrinsic( _inp, _outp )


#define WHILE_OUT(gubbins)	{ DWORD k; k=Port->Whileout;while((gubbins) && (--k > 0));\
							if(k==0) goto WhileoutError; }

/* the two pointers below are used to pass to the the ISRs the PPORT struct	*/
/* they are to use															*/
PPORT *gINT5_Port;
PPORT *gINT7_Port;

/* tmp variables for debug of ISR.											*/
BYTE volatile gINT5_Count = 0;
BYTE volatile gINT7_Count = 0;

 
/*--------------------------------------------------------------------------*/
/*	DetectPortType															*/
/*--------------------------------------------------------------------------*/
BOOL PP_DetectPortType( BYTE PortNumber, PPORT *Port ) {
	BYTE numports;
	BYTE int_line, dma_channel;

	Port->PortType = NULL_PORT_TYPE;

	numports = (BYTE)(( (*((BYTE far *)MAKELP(0x40, 0x11))) &0xC0 )>>6 );

	if ( ( numports < LPT1 ) || ( numports > LPT3 ) ) {
		return( FALSE );
	}
	if ( ( PortNumber < LPT1 ) || ( PortNumber > numports )	) {
		return( FALSE );
	}
	Port->BaseAddr = *((WORD far *)MAKELP(0x40, 6 + (2 * PortNumber)));

	if ( Port->BaseAddr == 0) {
		return( FALSE );
	}
	if ( (Port->BaseAddr != 0x378) 
		    && (Port->BaseAddr != 0x278) 
		    && (Port->BaseAddr != 0x3BC))	{
		return( FALSE );
	}

  	Port->PortType = TestPortType( Port->BaseAddr );
  
//	int_line = NO_PP_INT_AVAILABLE;
	int_line = FindIntLine( Port->BaseAddr );
	if ( int_line != NO_PP_INT_AVAILABLE ) {
		Port->PortType |= INT_PORT_TYPE;
		Port->IntLine = int_line;
//		dma_channel = NO_DMA_CHANNEL_AVAILABLE;
		dma_channel = FindDMAChannel( Port->BaseAddr );
		if ( dma_channel != NO_DMA_CHANNEL_AVAILABLE ) {
			Port->PortType |= DMA_PORT_TYPE;
			Port->DMAChannel = dma_channel;
		}
	}
	
 	Port->Timeout_mS = DEFAULT_TIMEOUT_MS;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	CreatePPort																*/
/*--------------------------------------------------------------------------*/
BOOL PP_CreatePPort( PPORT *Port, WORD BaseAddr, BOOL ECPPort, 
								  BYTE IntLine, BYTE DMAChannel ) {
	BYTE port_type;

	Port->BaseAddr = BaseAddr;
  	Port->IntLine = 0;
 	Port->DMAChannel = 0;
	
	port_type = COMPAT_PORT_TYPE;

	if ( (IntLine == 7) || (IntLine == 5) ) {
		port_type |= INT_PORT_TYPE;
	 	Port->IntLine = IntLine;
	}

	if ( ECPPort ) {
		port_type |= BIDIR_PORT_TYPE | ECP_PORT_TYPE;

		if ( port_type & (INT_PORT_TYPE | ECP_PORT_TYPE) ) {
			/* DMA is only available if we've got a usable INT line */
			if ( DMAChannel != 0 ) {
				port_type |= DMA_PORT_TYPE;
		 		Port->DMAChannel = DMAChannel;
			}
		}
	}

	Port->PortType = port_type;

 	Port->Timeout_mS = DEFAULT_TIMEOUT_MS;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	InitialisePort															*/
/*--------------------------------------------------------------------------*/
BOOL PP_InitialisePPort( PPORT *Port ) {
	WORD Base;
	double diff;
	DWORD start, stop;

	Base = Port->BaseAddr;

	Port->DataAddr = Base;
	Port->StatusAddr = Base + STATUS_OFFSET;
	Port->ControlAddr = Base + CONTROL_OFFSET;
	Port->ECRAddr = Base + ECP_ECR;
	Port->DataFIFOAddr = Base + ECP_D_FIFO;

 	Port->PrevControl = (BYTE)_inp( Port->ControlAddr );
	if ( !(Port->PrevControl & DirBit) ) { /* ie port was in FORWARD */
		Port->PrevData = (BYTE)_inp( Port->DataAddr );
	}

	if ( Port->PortType & ECP_PORT_TYPE ) {
		Port->PrevECR = (BYTE)_inp( Base + ECP_ECR );
		
		SetECRMode( Port, ECP_STANDARD_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_STANDARD_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_CNFG_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_CNFG_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_CNFG_MODE );
		Port->PrevCnfgA = (BYTE)_inp( Base + ECP_CNFG_A );
		Port->PrevCnfgB = (BYTE)_inp( Base + ECP_CNFG_B );

		/* set CnfgA and CnfgB to default values here if required 	*/
		/* ... probably safer to not touch them for now !!				*/
		SetECRMode( Port, ECP_STANDARD_MODE );
		if ( GetECRBit( Port, ECR_mode ) != ECP_STANDARD_MODE ) goto InitialiseError; 

		SetECRMode( Port, ECP_PS2_MODE );
		if ( GetECRBit( Port, ECR_mode ) !=  ECP_PS2_MODE ) goto InitialiseError; 
	}		

	Port->ControlShadow = (BYTE)(Port->PrevControl ^ CONTROL_INVERSION_MASK);
	SetCOMPATMode( Port );	
	ClearControlBit( Port, IntrEnable );

	/* determine port timing for timeout value  */
	Port->Whileout = 100000;					/* calibration value */

    start = GetTime();                                        
	WHILE_OUT( ((BYTE)_inp( Port->StatusAddr ) | 0xFF) );		/* fastest type of WHILE_OUT useage */

WhileoutError: /* not actually an error in this case !!! */
    stop = GetTime();                                        
	if ( start == stop ) goto InitialiseError;	/* Err, something went wrong !!!! */

	diff = ((stop - start) * 1000)/(DWORD)CLOCKS_PER_SEC;
    Port->Whileout = ((DWORD)Port->Timeout_mS * (DWORD)100000) / (DWORD)diff;
    /* whileout value is the number of loops needed to wait Timeout_mS seconds */

	Port->EnablePPInt = FALSE;
	if ( Port->PortType & INT_PORT_TYPE ) {

		/* clear down INT data structures					*/
		Port->PP_Callback = NULL;
		Port->DMA_Callback = NULL;
		Port->PP_Context = 0;
		Port->DMA_Context = 0;

 		Port->OldISRVector = DosGetIntVector( 8 + Port->IntLine );
		if ( !Hook_ISR( Port ) ) return( FALSE );
		Port->EnablePPInt = TRUE;
		if ( Port->PortType & DMA_PORT_TYPE ) {
#ifndef _WINDOWS
			if ( !DMA_Init( DMA_BUFFER_DOS_APP ) ) return( FALSE );
#else
			if ( !DMA_Init( DMA_BUFFER_PHYS ) ) return( FALSE );
#endif
		}
	}

 	return( TRUE );

InitialiseError:
	return( FALSE );
}


/*-----------------------------------------------------------------*/
/* ClosePort - Restores state of the parallel port to that before  */
/*-----------------------------------------------------------------*/
void PP_ClosePPort( PPORT *Port ) {
	WORD Base;

	Base = Port->BaseAddr;

	if ( Port->PortType & DMA_PORT_TYPE ) DMA_FreeBuffer( Port->DMAChannel );
	
	if ( Port->PortType & INT_PORT_TYPE ) Unhook_ISR( Port );

	if ( Port->PortType & ECP_PORT_TYPE ) {
		SetECRMode( Port, ECP_CNFG_MODE );
		_outp( Base + ECP_CNFG_A, Port->PrevCnfgA );
		_outp( Base + ECP_CNFG_B, Port->PrevCnfgB );
		_outp( Port->ECRAddr, Port->PrevECR );
	}
 
	if ( !(Port->PrevControl & DirBit) ) { /* ie port was in FORWARD */
		_outp( Port->DataAddr, Port->PrevData );
	}

	return;
}


#define VPP_WAIT_TIME	200
/*--------------------------------------------------------------------------*/
/*	SendVPP - Sends a number of data bytes to the  camera without any 		*/
/*				activity on the control lines.								*/
/*--------------------------------------------------------------------------*/
BOOL PP_SendVPP( PPORT *Port, BYTE Bytes, BYTE *Buffer ) {
	int i,j;

	for( i = 0; i < Bytes; i++ ) {
		for( j = 0; j < VPP_WAIT_TIME; j++ ) { /* this loop for delay purposes */
			PutData( Port, Buffer[ i ] );
		}
	}

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	GetHWInfo - Allows the retrieval of the hardware information			*/
/*--------------------------------------------------------------------------*/
BOOL PP_GetHWInfo( PPORT *Port, WORD *BaseAddr, BYTE *PortType,
				   BYTE *InterruptLine, BYTE *DMAChannel ) {

	if ( BaseAddr ) *BaseAddr = Port->BaseAddr;
	if ( PortType ) *PortType = Port->PortType;
	if ( InterruptLine ) *InterruptLine = Port->IntLine;
	if ( DMAChannel ) *DMAChannel = Port->DMAChannel;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetHWInfo - Sets the hardware information								*/
/*--------------------------------------------------------------------------*/
BOOL PP_SetHWInfo( PPORT *Port, WORD BaseAddr, BYTE PortType, /* if any of these are zero ignore */
				   BYTE InterruptLine, BYTE DMAChannel ) {

	if ( BaseAddr ) return( FALSE ); /* can't set this .... */

	if ( PortType ) Port->PortType = PortType;
	if ( InterruptLine ) Port->IntLine = InterruptLine;
	if ( DMAChannel ) Port->DMAChannel = DMAChannel;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	GetTimeout - Allows the retrieval of the current timeout value			*/
/*--------------------------------------------------------------------------*/
BOOL PP_GetTimeout( PPORT *Port, WORD *Timeout_mS ) {

	if ( Timeout_mS ) *Timeout_mS = Port->Timeout_mS;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetTimeout - Allows the setting of the timeout value					*/
/*--------------------------------------------------------------------------*/
BOOL PP_SetTimeout(  PPORT *Port, WORD Timeout_mS ) {

	if ( Timeout_mS < 10 ) Timeout_mS = 10;
	if ( Timeout_mS > 1000 ) Timeout_mS = 1000;
	
	Port->Timeout_mS = Timeout_mS;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetPPCallback - Sets the address of an PP Interrupt callback function	*/
/*					which will be called from the PP_ISR.					*/
/*--------------------------------------------------------------------------*/
 BOOL PP_SetPPCallback( PPORT *Port, void (*PP_Callback)( DWORD ContextParam ),
 						DWORD ContextParam ) {

	if ( !(Port->PortType & INT_PORT_TYPE) ) return( FALSE );

	if ( PP_Callback != NULL ) {
		Port->PP_Callback = PP_Callback;
		Port->PP_Context = ContextParam;
	} 
	else {
		Port->PP_Callback = NULL;
		Port->PP_Context = 0;
	}

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetDMACallback - Sets the address of an DMA Interrupt callback function	*/
/*					which will be called from the PP_ISR.					*/
/*--------------------------------------------------------------------------*/
 BOOL PP_SetDMACallback( PPORT *Port, void (*DMA_Callback)( DWORD ContextParam ),
 						 DWORD ContextParam ) {

	if ( !(Port->PortType & INT_PORT_TYPE) ) return( FALSE );

	if ( DMA_Callback != NULL ) {
		Port->DMA_Callback = DMA_Callback;
		Port->DMA_Context = ContextParam;
	}
	else {
		Port->DMA_Callback = NULL;
		Port->DMA_Context = 0;
  	}

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	EnablePPInterrupt - Enables/Disables the PP Interrupt 					*/
/*--------------------------------------------------------------------------*/
BOOL PP_EnablePPInterrupt( PPORT *Port, BOOL EnableInt ) {
	
	if ( !(Port->PortType & INT_PORT_TYPE) ) return( FALSE );

	if ( EnableInt == TRUE ) {
		Port->EnablePPInt = TRUE;
	}
	else {
		Port->EnablePPInt = FALSE;
	}
	return( TRUE );
}

 
/*--------------------------------------------------------------------------*/
/* INT5_ISR		- Int 5 ISR												 	*/
/*--------------------------------------------------------------------------*/
void __cdecl __interrupt __far PP_INT5_ISR( void ) 
{
	gINT5_Count++;
	if ( gINT5_Port ) 
    {
		if ( _inp( gINT5_Port->ECRAddr ) & ECR_dmaEn )		    /* ie DMA enabled .. */
        {
            if ( _inp( gINT5_Port->ECRAddr ) & ECR_serviceIntr )/* .. and TC reached */ 
            {
    			/* it is an ECP DMA TC interrupt	*/
    			if ( gINT5_Port->DMA_Callback ) gINT5_Port->DMA_Callback( gINT5_Port->DMA_Context );
    		}
            else 
            {
                /* something else generated the interrupt       */
                /* ... just ignore it.                          */
            }
        }
 	    else
        {
            if ( _inp( gINT5_Port->ControlAddr ) & IntrEnable )
            {
               /* it is a Std PP interrupt 	*/
        		if ( gINT5_Port->PP_Callback ) gINT5_Port->PP_Callback( gINT5_Port->PP_Context );
            }
            else 
            {
                /* something else generated the interrupt       */
                /* ... just ignore it.                          */
            }
        }
    }
  
    _outp( PIC_CTRL, EOI_CODE );             
	return;
}



/*--------------------------------------------------------------------------*/
/* INT7_ISR		- Int 7 ISR												 	*/
/*--------------------------------------------------------------------------*/
void __cdecl __interrupt __far PP_INT7_ISR( void )
{
    gINT7_Count++;

	WIGGLE_LPT2( nAutoFd );

	if ( gINT7_Port ) 
    {
        if ( _inp( gINT7_Port->ECRAddr ) & ECR_dmaEn)			/* ie DMA enabled .. */
        {
            if ( _inp( gINT7_Port->ECRAddr ) & ECR_serviceIntr )/* .. and TC reached */ 
            {
				/* it is an ECP DMA TC interrupt	*/
				if ( gINT7_Port->DMA_Callback ) gINT7_Port->DMA_Callback( gINT7_Port->DMA_Context );
			}
            else
            {
                /* something else generated the interrupt       */
                /* ... just ignore it.                          */
            }
        }
        else
        {
            if ( _inp( gINT7_Port->ControlAddr ) & IntrEnable )
            {
                /* it is a Std PP interrupt 	*/
                if ( gINT7_Port->PP_Callback ) gINT7_Port->PP_Callback( gINT7_Port->PP_Context );
            }
            else 
            {
                /* something else generated the interrupt       */
                /* ... just ignore it.                          */
            }
        }
	}
	WIGGLE_LPT2( nAutoFd );
	
	_outp( PIC_CTRL, EOI_CODE );             
	return;
}


/*--------------------------------------------------------------------------*/
/*	SetTransferMode - Negotiates from compatibility mode into the requested */
/*					  1284 standard mode.									*/
/*--------------------------------------------------------------------------*/
BOOL PP_SetTransferMode( PPORT *Port, BYTE NewMode ) {
	BYTE request_id;

	if ( Port->TransferMode != COMPAT_TRANSFER ) goto ModeNotChanged;
	
	request_id = (BYTE)(NewMode & REQUEST_DEVICE_ID);
	
	switch( (NewMode & ~REQUEST_DEVICE_ID) ) {

	case ECP_TRANSFER:
		if ( !Negotiate2SetupPhase( Port, (BYTE)(EXTEN_ECP_MODE | request_id), 0 ) ) goto ModeNotChanged;
		if ( !ECPSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	case NIBBLE_TRANSFER:
		if ( !Negotiate2SetupPhase( Port, (BYTE)(EXTEN_NIBBLE_MODE | request_id), 0 ) ) goto ModeNotChanged;
		if ( !NibbleSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	default:
		goto ModeNotChanged;
	}
	return( TRUE );

ModeNotChanged:
	return( FALSE );
}

/*-----------------------------------------------------------------*/
/* EndTransferMode												  */
/*-----------------------------------------------------------------*/
BOOL PP_EndTransferMode( PPORT *Port ) {
	if ( Port->TransferMode == COMPAT_TRANSFER ) return( TRUE );

	switch( Port->TransferMode ) {

	case ECP_TRANSFER:
		if ( Port->Direction == REVERSE_DIR ) {
			if ( !PP_Reverse2Forward( Port ) ) goto ModeNotChanged;
		}
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	case NIBBLE_TRANSFER:
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	default:
		goto ModeNotChanged;
	}

	return( TRUE );

ModeNotChanged:
	return( FALSE );

DoImmediateTermination:
	Immediate1284Termination( Port );
	return( FALSE );
}



/*------------------------------------------------------------------*/
/* SetUploadMode														*/
/*------------------------------------------------------------------*/
BOOL PP_SetUploadMode( PPORT *Port, BYTE UploadMode, BYTE ModeFlags ) {

	if ( Port->TransferMode != COMPAT_TRANSFER ) goto ModeNotChanged;

	if ( !Negotiate2SetupPhase( Port, REQUEST_LINK, (BYTE)(UploadMode | ModeFlags) ) )
		goto ModeNotChanged;

	switch( UploadMode ) {
	
	case ECP_UPLOAD:  	/* since this is almost the same as ECP		*/
					  	/* it can share the ECP Setup stuff.		*/
		if ( !ECPSetupPhase( Port ) ) goto ModeNotChanged;
		if ( !PP_Forward2Reverse( Port ) ) {
			Valid1284Termination( Port );
			goto ModeNotChanged;
		}
		break;

	case NIBBLE_UPLOAD:
		if ( !NibbleSetupPhase( Port ) ) goto ModeNotChanged;
		break;

	default:
		goto ModeNotChanged;
	}

	Port->TransferMode = UploadMode;
	return( TRUE );

ModeNotChanged:
	return( FALSE );
}



/*-----------------------------------------------------------------*/
/* EndUploadMode													  */
/*-----------------------------------------------------------------*/
BOOL PP_EndUploadMode( PPORT *Port ) {
	if ( Port->TransferMode == COMPAT_TRANSFER ) return( TRUE );

	switch( (Port->TransferMode & ~RLE_DATA_FLAG) ) {

	case ECP_UPLOAD:
	case BIDIR_UPLOAD:
		if ( Port->Direction == REVERSE_DIR )
			if ( !PP_Reverse2Forward( Port ) ) goto DoImmediateTermination;
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	case NIBBLE_UPLOAD:
		if ( !Valid1284Termination( Port ) ) goto DoImmediateTermination;
		break;

	default:
		goto ModeNotChanged;
	}
	return( TRUE );

ModeNotChanged:
	return( FALSE );

DoImmediateTermination:
	Immediate1284Termination( Port );
	return( FALSE );
}

/*-----------------------------------------------------------------*/
/* Attempt1284Negotiation										  */
/*-----------------------------------------------------------------*/
BOOL PP_Attempt1284Negotiation( PPORT *Port ) {
	
	DWORD tmp_whileout;
	
 	PutData( Port, 0 );	/* ie request reverse nibble mode */

	SetControlBit( Port, Active_1284 );
	ClearControlBit( Port, nAutoFd );
	
	/* frig the WHILE_OUT mechanism to give a very short timeout */
	tmp_whileout = Port->Whileout;
	Port->Whileout = 50;

	WHILE_OUT( GetStatusBit( Port, nAck ) );			/* check nAck LO ...	*/
	WHILE_OUT( !((GetStatus( Port ) & (PError|nFault|XFlag))
					== (PError|nFault|XFlag))  );			/* ... and these lines go HI 	*/

	Port->Whileout = tmp_whileout;
	/* if we got here then the dongle has responded */

	ClearControlBit( Port, nStrobe );
	
	SetControlBit( Port, nInit );	/* do this to ensure that the dongle does reset	*/
	ClearControlBit( Port, Active_1284 );			
   
	/* setting Active_1284 high here will cause the dongle to abort the */
	/* negotiation while it is waiting for event (4) to occur.			*/
	/* it jumps directly to its Immmediate1284Termination routine.		*/

	/* wait for peripheral to see the termination .. */
//	WHILE_OUT( GetStatusBit( Port, XFlag ) );

	SetCOMPATMode( Port );

	/* replace correct WHILE_OUT counts */
	Port->Whileout = tmp_whileout;
	return( TRUE );

WhileoutError:
	 SetCOMPATMode( Port );

	/* replace correct WHILE_OUT counts */
	Port->Whileout = tmp_whileout;
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	Forward2Reverse															*/
/*--------------------------------------------------------------------------*/
BOOL PP_Forward2Reverse( PPORT *Port ) {

	SetControlBit( Port, DirBit );		/* tri-state data bits */
	ClearControlBit( Port, HostAck );

	ClearControlBit( Port, nReverseRequest );
	WHILE_OUT( GetStatusBit( Port, nAckReverse ) );

	Port->Direction = REVERSE_DIR;
	return( TRUE );

WhileoutError:
	SetControlBit( Port, nReverseRequest );
	ClearControlBit( Port, DirBit );		/* drive data bits */
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	PP_Reverse2Forward															*/
/*--------------------------------------------------------------------------*/
BOOL PP_Reverse2Forward( PPORT *Port ) {

	SetControlBit( Port, nReverseRequest );
	WHILE_OUT( !GetStatusBit( Port, nAckReverse ) );
	ClearControlBit( Port, DirBit );		/* drive data bits */

	Port->Direction = FORWARD_DIR;
	return( TRUE );

WhileoutError:
	ClearControlBit( Port, nReverseRequest );
	SetControlBit( Port, DirBit );		/* tri-state data bits */
	return( FALSE );
}




/*--------------------------------------------------------------------------*/
/*	ECPWriteBuffer															*/
/*--------------------------------------------------------------------------*/
BOOL PP_ECPWriteBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	DWORD i;
	WORD j;
	BYTE __far *ptr;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 
		   || (Port->TransferMode != ECP_TRANSFER) 
		   || (Port->Direction != FORWARD_DIR) ) {
		
		return( FALSE );	
	}

	EmptyFIFO( Port );

	SetControlBit( Port, HostAck );	  				/* NOT what I'd expect to do !!! */
	SetControlBit( Port, HostClk ); 	 		   
	SetECRMode( Port, ECP_ECP_MODE );				

	ptr = Data;

	for( i = 0; i < (Bytes/ECP_FIFO_SIZE); i++ ) {	

		WHILE_OUT( !GetECRBit( Port, ECR_empty ) );  	/* wait for FIFO to empty */
		
		/* FIFO is now empty so safe to write at least ECP_FIFO_SIZE BYTE s */
		for( j = 0; j < ECP_FIFO_SIZE; j++ ) {
			_outp( Port->DataFIFOAddr, *ptr++);
		}
	}

	WHILE_OUT( !(GetECRBit( Port, ECR_empty )) ); 	
	for( j = 0; j < (Bytes%ECP_FIFO_SIZE); j++ ) {
		_outp( Port->DataFIFOAddr, *ptr++);
	}

	/* wait for last BYTE to be sent */
//	WHILE_OUT( !(GetECRBit( Port, ECR_empty )) ); 	

	/* check four times in case we're on a dodge ALi Motherboard .... */
	WHILE_OUT( !(GetECRBit( Port, ECR_empty )) 
				|| !(GetECRBit( Port, ECR_empty ))
				|| !(GetECRBit( Port, ECR_empty ))
				|| !(GetECRBit( Port, ECR_empty ))	); 	

	SetECRMode( Port, ECP_PS2_MODE );				

	return( TRUE );

WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	EmptyFIFO( Port );
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	PP_ECPReadBuffer														*/
/*--------------------------------------------------------------------------*/
BOOL PP_ECPReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	WORD i, loops;
	WORD j;
	BYTE __far *ptr;
	BYTE __far *end_ptr;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 	 
		   || ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD)) 
		   || (Port->Direction != REVERSE_DIR) ) {
		return( FALSE );	
	}
	ptr = Data;
	end_ptr = ptr + Bytes;

	if ( Bytes < ECP_FIFO_SIZE*4 ) goto FinishManual;

	ClearControlBit( Port, HostAck );	  				

	/* ECP hardware only detects the HI->LO transition on PeriphClk, therefore
	   it will have missed this transition for the first BYTE ( which the dongle
	   sends straight after doing PP_Forward2Reverse() )
	   So here we manually read the first BYTE of data before enabling ECP hardware */ 
	WHILE_OUT( GetStatusBit( Port, PeriphClk ) );  	
	*ptr++ = GetData( Port );
	Bytes--;

	/* acknowledge for this BYTE */
	SetControlBit( Port, HostAck );	  				
	WHILE_OUT( !GetStatusBit( Port, PeriphClk ) ); 
	/* note - switching on ECP hardware will complete handshake for this BYTE by 	*/
	/* taking HostAck LO again.														*/

	/* OK we've got the first BYTE ... now let the ECP hardware do its thing */
	SetControlBit( Port, HostClk ); 	 			
	SetECRMode( Port, ECP_ECP_MODE );				

 	loops = (Bytes/ECP_FIFO_SIZE)-1;
	if ( Port->BaseAddr == LPT1_BASE_ADDR )	{
	 	for( i = 0; i < loops; i++ ) {	
			while( !((BYTE)_inp(ECR_LPT1) & ECR_full) );   /* wait for FIFO to fill */
			for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT1 );
		}
		/* switch off automatic filling of the FIFO		 	*/
		ClearControlBit( Port, HostAck );	  				
	 	/* read any remaining bytes out the FIFO 			*/
		while( (!((BYTE)_inp(ECR_LPT1) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT1 );
	}
	else {
		if ( Port->BaseAddr == LPT2_BASE_ADDR ) {
		 	for( i = 0; i < loops; i++ ) {	
				WHILE_OUT( !((BYTE)_inp(ECR_LPT2) & ECR_full) );   /* wait for FIFO to fill */
				for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT2 );
			}
			ClearControlBit( Port, HostAck );	  				
			while( (!((BYTE)_inp(ECR_LPT2) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT2 );
		}
		else {
			if ( Port->BaseAddr == LPT3_BASE_ADDR )	 {
			 	for( i = 0; i < loops; i++ ) {	
					WHILE_OUT( !((BYTE)_inp( ECR_LPT3 ) & ECR_full) );   /* wait for FIFO to fill */
					for( j = 0; j < ECP_FIFO_SIZE; j++ ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT3 );
				}
				ClearControlBit( Port, HostAck );	  				
				while( (!((BYTE)_inp(ECR_LPT3) & ECR_empty)) && ( ptr != end_ptr ) ) *ptr++ = (BYTE)_inp( DATA_FIFO_LPT3 );
			}
		}
	}

	while( !GetECRBit( Port, ECR_empty ) && ( ptr != end_ptr ) ) {
  		*ptr++ = (BYTE)_inp( Port->DataFIFOAddr );
	}

 	SetECRMode( Port, ECP_PS2_MODE );				
//	SetControlBit( Port, HostAck );

	/* now manually pull through any remaining bytes 	*/

FinishManual:
	PP_SimECPReadBuffer( Port, ptr, (end_ptr - ptr) );
	 
	return( TRUE );

WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	SimECPWriteBuffer														*/
/*--------------------------------------------------------------------------*/
BOOL PP_SimECPWriteBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	DWORD i;
	BYTE __far *ptr;

	if ( (Port->TransferMode != ECP_TRANSFER)
			|| (Port->Direction != FORWARD_DIR) ) {
	   return( FALSE );	
	}

	SetControlBit( Port, HostAck );	  	/* this is a data transfer	*/

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {	
		PutData( Port, *ptr++ );

		ClearControlBit( Port, HostClk );
		WHILE_OUT( !GetStatusBit( Port, PeriphAck ) );  	

		SetControlBit( Port, HostClk );
		WHILE_OUT( GetStatusBit( Port, PeriphAck ) );  	
	}

	return( TRUE );

WhileoutError:
	SetControlBit( Port, HostClk );
	return( FALSE );
}

/*--------------------------------------------------------------------------*/
/*	PP_SimECPReadBuffer														*/
/*--------------------------------------------------------------------------*/
BOOL PP_SimECPReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	DWORD i;
	BYTE __far *ptr;

	if ( ( !(Port->PortType & BIDIR_PORT_TYPE))
			|| ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD))
			|| (Port->Direction != REVERSE_DIR) ) {
	   return( FALSE );	
	}

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {	
		ClearControlBit( Port, HostAck );
 		WHILE_OUT( GetStatusBit( Port, PeriphClk ) );

		*ptr++ = GetData( Port );		/* read data on rising edge of PeriphClk */

		SetControlBit( Port, HostAck );
		WHILE_OUT( !GetStatusBit( Port, PeriphClk ) );
  	}
		 
	return( TRUE );

WhileoutError:
	ClearControlBit( Port, HostAck );
	return( FALSE );
}

#define UPPER_NIBBLE	0xF0
#define LOWER_NIBBLE	0x0F


BYTE UnscrambleNibble( BYTE Data ) {

	Data = (BYTE)(Data & NIBBLE_DATA_BITS);

	if ( Data & Busy ) {
		return( (BYTE)(((Busy>>1) | (Data & 0x78))>>3) );
	}
	else {
		return( (BYTE)(Data>>3) );
	   	/* busy (bit 7 ) was not set, so no action is required ( bit 6 is already zero ) */
	}
}



/*--------------------------------------------------------------------------*/
/*	NibbleReadBuffer														*/
/*--------------------------------------------------------------------------*/
BOOL PP_NibbleReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes ) {

	DWORD i;
	BYTE __far *ptr;
	WORD tmp;

	if ( Port->TransferMode != NIBBLE_TRANSFER ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
 //		WHILE_OUT( GetStatusBit( Port, nDataAvail ) );
		
		ClearControlBit( Port, HostBusy );
		WHILE_OUT( GetStatusBit( Port, PtrClk ) );
		tmp = UnscrambleNibble( GetStatus( Port ) );
		
		/* now invert and shift BUSY bit */

		SetControlBit( Port, HostBusy );
		WHILE_OUT( !GetStatusBit( Port, PtrClk ) );

		ClearControlBit( Port, HostBusy );
		WHILE_OUT( GetStatusBit( Port, PtrClk ) );
		*ptr++ = (BYTE )(tmp | (UnscrambleNibble(GetStatus( Port ))<<4) );

		SetControlBit( Port, HostBusy );
		WHILE_OUT( !GetStatusBit( Port, PtrClk ) );
	}
	return( TRUE );

WhileoutError:

	SetControlBit( Port, HostBusy );
	return( FALSE );
}





/****************************************************************************/
/*																			*/
/*	Special PPC Upload Functions	 										*/
/*																			*/
/*																			*/
/****************************************************************************/

/*--------------------------------------------------------------------------*/
/*	NibbleUpload															*/
/*--------------------------------------------------------------------------*/
BOOL PP_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
    BOOL rtn;
    
	rtn = FALSE;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	if ( Port->BaseAddr == LPT1_BASE_ADDR )
		rtn = LPT1_NibbleUpload( Port, Data, Bytes );
	else
		if ( Port->BaseAddr == LPT2_BASE_ADDR )
			rtn = LPT2_NibbleUpload( Port, Data, Bytes );
		else
			if ( Port->BaseAddr == LPT3_BASE_ADDR )
				rtn = LPT3_NibbleUpload( Port, Data, Bytes );

	return( rtn );
}


/*--------------------------------------------------------------------------*/
/*	ECPUpload																*/
/*--------------------------------------------------------------------------*/
BOOL PP_ECPUpload( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	WORD fifo_unload;

	if ( (!(Port->PortType & ECP_PORT_TYPE) ) 
			|| (Port->TransferMode != ECP_UPLOAD) ) {
		return( FALSE );
	}

	/* read first byte and switch port to ECP mode 	*/
	if ( !PP_StartECPUpload( Port, Data ) ) return( FALSE );
	Data++;
	Bytes--;

	fifo_unload	= ((Bytes / ECP_FIFO_SIZE) * ECP_FIFO_SIZE) - ECP_FIFO_SIZE;

	/* now read the data via the FIFO				*/
	if ( !PP_UnloadFIFO( Port, Data, fifo_unload ) ) goto ECPUploadError;		
	Data += fifo_unload;
	Bytes -= fifo_unload;

	/* stop the periperal from sending any more data */
	ClearControlBit( Port, HostAck );	  				

	/* empty any bytes remaining in the FIFO		*/
	if ( !PP_EmptyFIFO( Port, Data, &fifo_unload ) ) goto ECPUploadError;
	Data += fifo_unload;
	Bytes -= fifo_unload;

	/* switch off the ECP hardware 					*/
	if ( !PP_FinishECPUpload( Port ) ) goto ECPUploadError;

	/* manually read any bytes still required		*/
	if ( !PP_SimECPReadBuffer( Port, Data, Bytes ) ) goto ECPUploadError;
 
	return( TRUE );

ECPUploadError:
	return( TRUE );
}



/****************************************************************************/
/*																			*/
/*	Low level upload functions												*/
/*																			*/
/****************************************************************************/

/*--------------------------------------------------------------------------*/
/*	StartECPUpload 															*/
/*--------------------------------------------------------------------------*/
BOOL PP_StartECPUpload( PPORT *Port, BYTE __far* Data ) {

	if ( Port->TransferMode != ECP_UPLOAD )	return( FALSE );	

	ClearControlBit( Port, HostAck );	  				

	/* ECP hardware only detects the HI->LO transition on PeriphClk, therefore
	   it will have missed this transition for the first BYTE (which the dongle
	   sends straight after doing PP_Forward2Reverse())
	   So here we manually read the first BYTE of data before enabling ECP hardware */ 
	WHILE_OUT( GetStatusBit( Port, PeriphClk ) );  	
	*Data = GetData( Port );

	/* acknowledge for this BYTE */
	SetControlBit( Port, HostAck );	  				
	WHILE_OUT( !GetStatusBit( Port, PeriphClk ) ); 
	/* note - switching on ECP hardware will complete handshake for this BYTE by 	*/
	/* taking HostAck LO again.														*/

	/* OK we've got the first BYTE ... now let the ECP hardware do its thing */
	SetControlBit( Port, HostClk ); 	 			
	SetECRMode( Port, ECP_ECP_MODE );				

	return( TRUE );

 WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	FinishUpload															*/
/*--------------------------------------------------------------------------*/
BOOL PP_FinishECPUpload( PPORT *Port ) {

	if ( Port->TransferMode != ECP_UPLOAD ) return( FALSE );	

	ClearControlBit( Port, HostAck );	  				
  	SetECRMode( Port, ECP_PS2_MODE );				

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	UnloadFIFO																*/
/*--------------------------------------------------------------------------*/
BOOL PP_UnloadFIFO( PPORT *Port, BYTE __far* Data, WORD Bytes ) {
	WORD i, j, loops, remainder;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 	 
		   || ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD)) 
		   || (Port->Direction != REVERSE_DIR) ) {
		return( FALSE );	
	}

 	loops = Bytes/ECP_FIFO_SIZE;
	remainder = Bytes%ECP_FIFO_SIZE;
	
	switch( Port->BaseAddr ) {
	case LPT1_BASE_ADDR:
	 	for( i = 0; i < loops; i++ ) {	
			WHILE_OUT( !((BYTE)_inp(ECR_LPT1) & ECR_full) );   /* wait for FIFO to fill */
			for( j = 0; j < ECP_FIFO_SIZE; j++ ) *Data++ = (BYTE)_inp( DATA_FIFO_LPT1 );
		}
		if ( remainder ) {
			for( i = 0; i < remainder; i++ ) {
				WHILE_OUT( ((BYTE)_inp(ECR_LPT1) & ECR_empty) );
				*Data++ = (BYTE)_inp( DATA_FIFO_LPT1 );
			}
		}
		break;
	case LPT2_BASE_ADDR:
		for( i = 0; i < loops; i++ ) {	
			WHILE_OUT( !((BYTE)_inp(ECR_LPT2) & ECR_full) );   /* wait for FIFO to fill */
			for( j = 0; j < ECP_FIFO_SIZE; j++ ) *Data++ = (BYTE)_inp( DATA_FIFO_LPT2 );
		}
		if ( remainder ) {
			for( i = 0; i < remainder; i++ ) {
				WHILE_OUT( ((BYTE)_inp(ECR_LPT2) & ECR_empty) );
				*Data++ = (BYTE)_inp( DATA_FIFO_LPT2 );
			}
		}
		break;
	case LPT3_BASE_ADDR:
	 	for( i = 0; i < loops; i++ ) {	
			WHILE_OUT( !((BYTE)_inp( ECR_LPT3 ) & ECR_full) );   /* wait for FIFO to fill */
			for( j = 0; j < ECP_FIFO_SIZE; j++ ) *Data++ = (BYTE)_inp( DATA_FIFO_LPT3 );
		}
		if ( remainder ) {
			for( i = 0; i < remainder; i++ ) {
				WHILE_OUT( ((BYTE)_inp(ECR_LPT3) & ECR_empty) );
				*Data++ = (BYTE)_inp( DATA_FIFO_LPT3 );
			}
		}
		break;
	}

	return( TRUE );

WhileoutError:
	SetECRMode( Port, ECP_PS2_MODE );				
	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	EmptyFIFO																*/
/*--------------------------------------------------------------------------*/
BOOL PP_EmptyFIFO( PPORT *Port, BYTE __far* Data, WORD *Bytes ) {
	WORD bytes;

	if ( (!(Port->PortType & ECP_PORT_TYPE)) 	 
		   || ((Port->TransferMode != ECP_TRANSFER) && (Port->TransferMode != ECP_UPLOAD)) 
		   || (Port->Direction != REVERSE_DIR) ) {
		return( FALSE );	
	}
		
	bytes = 0;

	switch( Port->BaseAddr ) {
	case LPT1_BASE_ADDR:
		while( !((BYTE)_inp(ECR_LPT1) & ECR_empty) ) {
			*Data++ = (BYTE)_inp( DATA_FIFO_LPT1 );
			bytes++;
		}
		break;
	case LPT2_BASE_ADDR:
		while( !((BYTE)_inp(ECR_LPT2) & ECR_empty) ) {
			*Data++ = (BYTE)_inp( DATA_FIFO_LPT2 );
			bytes++;
		}
		break;
	case LPT3_BASE_ADDR:
		while( !((BYTE)_inp(ECR_LPT3) & ECR_empty) ) {
			*Data++ = (BYTE)_inp( DATA_FIFO_LPT3 );
			bytes++;
		}
		break;
	}

	*Bytes = bytes;

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	SetDMABuffer															*/
/*--------------------------------------------------------------------------*/
BOOL PP_SetDMABuffer( PPORT *Port, BYTE __far* Data, DWORD PhysicalAddr, WORD BuffSize ) {
	BOOL rtn;
	BYTE tmp;

	rtn = DMA_SetBuffer( Port->DMAChannel, Data, PhysicalAddr, BuffSize );
	if ( !rtn ) return( FALSE );
	
	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	StartDMAUploadFIFO														*/
/*--------------------------------------------------------------------------*/
BOOL PP_StartDMAUnloadFIFO( PPORT *Port, WORD Bytes ) {
	BOOL rtn;
	BYTE tmp;

	rtn = DMA_StartTransfer( Port->DMAChannel, (Bytes-1) );
	if ( !rtn ) return( FALSE );
	
	/* set ECR_dmaEn and clear ECR_serviceIntr bits */
	tmp = _inp( Port->ECRAddr );
	tmp = (tmp & ~ECR_serviceIntr) | ECR_nErrIntrEn | ECR_dmaEn; 
	_outp( Port->ECRAddr, tmp );

	/* start ECP transfer */

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	CheckDMAComplete														*/
/*--------------------------------------------------------------------------*/
BOOL PP_CheckDMAComplete( PPORT *Port ) {

	if ( DMA_CheckTCBit( Port->DMAChannel ) ) {
		DMA_EndTransfer( Port->DMAChannel ); 
    	return( TRUE );
    }
    else 
    	return( FALSE );
}                            


/*--------------------------------------------------------------------------*/
/*	StopDMAUnloadFIFO														*/
/*--------------------------------------------------------------------------*/
BOOL PP_StopDMAUnloadFIFO( PPORT *Port ) {
	BYTE tmp;

	/* clear ECR_dmaEn and set ECR_serviceIntr bits */
	tmp = _inp( Port->ECRAddr );
	tmp = (tmp & ~ECR_dmaEn) | ECR_nErrIntrEn | ECR_serviceIntr; 
	_outp( Port->ECRAddr, tmp );

	return( TRUE );
}


/*--------------------------------------------------------------------------*/
/*	AbortDMAUnload															*/
/*--------------------------------------------------------------------------*/
BOOL PP_AbortDMAUnload( PPORT *Port ) {

	PP_StopDMAUnloadFIFO( Port );

	DMA_AbortTransfer( Port->DMAChannel );

	return( TRUE );
}







/****************************************************************************/
/*																			*/
/*	Private functions of PPORT Module below here 							*/
/*																			*/
/****************************************************************************/


void SetCOMPATMode( PPORT *Port ) {

	SetControlBit( Port, (nStrobe|nAutoFd|nInit) );
	ClearControlBit( Port, nSelectIn|DirBit );		
	
	PutData( Port, 0xFF ); 	/* just in case peripheral is still driving lines */

	Port->TransferMode = COMPAT_TRANSFER;
	Port->Direction = FORWARD_DIR;

	return;
}




/*--------------------------------------------------------------------------*/
/*	Negotiate2SetupPhase													*/
/*--------------------------------------------------------------------------*/
int Negotiate2SetupPhase( PPORT *Port, BYTE ExtensibilityByte, BYTE LinkMode ) {

 	ClearControlBit( Port, IntrEnable );
 
	SetDataDirection( Port, FORWARD_DIR );
	PutData( Port, ExtensibilityByte );

	SetControlBit( Port, Active_1284 );
	ClearControlBit( Port, nAutoFd );

 	WHILE_OUT( GetStatusBit( Port, nAck ) );			/* check nAck LO ...	*/
	WHILE_OUT( !((GetStatus( Port ) & (PError|nFault|XFlag))
				== (PError|nFault|XFlag))  );			/* ... and these lines go HI 	*/

	/* disable the PP interrupt line here */
//	ClearControlBit( Port, IntrEnable );
 
	/* ok we have got a 1284 peripheral attached */

	/* the following LO going pulse on nStrobe is now ignored by the 		*/
	/* periperal which will read the ExtensiblityByte as soon as it sees	*/
	/* Active_1284 going HI.												*/
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );
	ClearControlBit( Port, nStrobe );

 	SetControlBit( Port, (nAutoFd | nStrobe) );		   /* min 0.5uS LO pulse required 	*/
  
  	WHILE_OUT( !GetStatusBit( Port, nAck ) ); 					//  -- (event 6)		
	
//	ExtensibilityByte &= ~REQUEST_DEVICE_ID;		// clear Device ID Request bit

	if ( ExtensibilityByte == EXTEN_NIBBLE_MODE ) {
		if ( GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}
	else {
		if ( !GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}

	if ( ExtensibilityByte & REQUEST_LINK ) {
		/* an Extensiblity Link was requested */

		PutData( Port, LinkMode );								
					
		ClearControlBit( Port, nStrobe);							//  -- (event 51)
	  	WHILE_OUT( GetStatusBit( Port, nAck ) ); 							

		SetControlBit( Port, (nAutoFd | nStrobe) );		   			//  -- (event 53)
	  	WHILE_OUT( !GetStatusBit( Port, nAck ) ); 				   		

		if ( !GetStatusBit( Port, XFlag) ) goto ModeNotSupported;
	}
  	/* negotiation phase complete */

	return( TRUE );	

WhileoutError:
	SetCOMPATMode( Port );
 	return( FALSE );

ModeNotSupported:
	SetCOMPATMode( Port );
 	return( FALSE );
}


/*--------------------------------------------------------------------------*/
/*	Valid1284Termination													*/
/*--------------------------------------------------------------------------*/
int Valid1284Termination( PPORT *Port ) {

	SetControlBit( Port, nInit );	/* do this to ensure that the dongle 	*/
									/* doesn't get a reset.					*/
									/* Reset is  nInit = Active_1284 = LO	*/

	SetControlBit( Port, nAutoFd );

	ClearControlBit( Port, Active_1284 );			

//
//	WHILE_OUT( GetStatusBit( Port, (Busy|nFault) ) 	
//						!= (Busy|nFault) );			/* 		-- (event 23)	*/

                                                
	WHILE_OUT( GetStatusBit( Port, nAck ) );		/*		-- (event 24)	*/ 
	ClearControlBit( Port, nAutoFd );				/*		-- (event 25)	*/
                                                
                                                
	WHILE_OUT( !GetStatusBit( Port, nAck ) );		/* 		-- (event 27)	*/

 	if ( Port->EnablePPInt ) SetControlBit( Port, IntrEnable );

	SetControlBit( Port, nAutoFd );

	SetCOMPATMode( Port );
    
    return( TRUE );

WhileoutError:
	Immediate1284Termination( Port );
	return( FALSE );
}




/*--------------------------------------------------------------------------*/
/*	Immmediate1284Termination												*/
/*--------------------------------------------------------------------------*/
void Immediate1284Termination( PPORT *Port ) {

	SetControlBit( Port, nInit );	/* make sure dongle will not be reset	*/
	ClearControlBit( Port, Active_1284 );

	/* wait for peripheral to see the termination .. */
	WHILE_OUT( GetStatusBit( Port, XFlag ) );

	if ( Port->EnablePPInt ) SetControlBit( Port, IntrEnable );

	SetCOMPATMode( Port );
	return;

WhileoutError:
   	SetCOMPATMode( Port );
	return;
}




											
/*--------------------------------------------------------------------------*/
/*	ECPSetupPhase															*/
/*--------------------------------------------------------------------------*/
int  ECPSetupPhase( PPORT *Port ) {
 
 	/* Assertion - negotiation phase complete before this functions is 		*/
 	/* called ... now do ECP setup phase.									*/

	ClearControlBit( Port, HostAck );

	WHILE_OUT( !GetStatusBit( Port, nAckReverse ) );

	EmptyFIFO( Port );

	Port->TransferMode = ECP_TRANSFER;
	Port->Direction = FORWARD_DIR;
 	return( TRUE );

WhileoutError:
	SetCOMPATMode( Port );
 	return( FALSE );
}

 

/*--------------------------------------------------------------------------*/
/*	AttemptForward2Reverse	  												*/
/*--------------------------------------------------------------------------*/
int AttemptForward2Reverse( PPORT *Port ) {
	static int WaitingForReverseAck = FALSE;

	if ( Port->Direction != FORWARD_DIR ) {
		WaitingForReverseAck = FALSE;
		return( FALSE );
	}
	
	if ( !WaitingForReverseAck ) {
		SetControlBit( Port, DirBit );		/* tri-state data bits */
		ClearControlBit( Port, HostAck );
		ClearControlBit( Port, nReverseRequest );

		/* wait a bit here */
		ClearControlBit( Port, nReverseRequest );
		ClearControlBit( Port, nReverseRequest );
		ClearControlBit( Port, nReverseRequest );

		WaitingForReverseAck = TRUE;	
	}

	if ( !GetStatusBit( Port, nAckReverse ) ) { 	/* peripheral has acknowledged */
		Port->Direction = REVERSE_DIR;
		WaitingForReverseAck = FALSE;
		return( TRUE );
	}	
	else {
		return( FALSE );							/* peripheral has not acknowledged */
	}
}


void SetECRMode( PPORT *Port, BYTE Mode ) {
	_outp( Port->ECRAddr, ( Mode | ECR_nErrIntrEn | ECR_serviceIntr ) );
}

BYTE GetECRBit( PPORT *Port, BYTE BitMask ) {
	BYTE tmp;
	tmp = (BYTE)_inp( Port->BaseAddr + ECP_ECR );
	return( (BYTE)(tmp & BitMask) );
}

void EmptyFIFO( PPORT *Port ) {
	if ( Port->PortType & ECP_PORT_TYPE ) {
		WHILE_OUT( !GetECRBit( Port, ECR_empty ) ) (BYTE)_inp( Port->DataFIFOAddr );
	}
WhileoutError:
	return;
}



/*--------------------------------------------------------------------------*/
/*	NibbleSetupPhase						 								*/
/*--------------------------------------------------------------------------*/
int  NibbleSetupPhase( PPORT *Port ) {
 
 	/* Assertion - negotiation phase complete before this functions is 		*/
 	/* called ... now do Nibble setup phase. 								*/

	SetDataDirection( Port, REVERSE_DIR );

//	ClearControlBit( Port, HostBusy );

	Port->TransferMode = NIBBLE_TRANSFER;
	Port->Direction = REVERSE_DIR;

 	return( TRUE );
}



/* ====================================================================	*/
/*																		*/
/* General routines														*/
/*																		*/
/* ====================================================================	*/

/*--------------------------------------------------------------------------*/
/*	TestPortType															*/
/*--------------------------------------------------------------------------*/
BYTE TestPortType( WORD BaseAddr ) {
	BYTE port_type;

	port_type = COMPAT_PORT_TYPE;
	
	if ( TestGenericECP( BaseAddr ) ) 
		port_type |= (ECP_PORT_TYPE | BIDIR_PORT_TYPE);

	return( port_type );
}


/*--------------------------------------------------------------------------*/
/*	TestGenericECP															*/
/*--------------------------------------------------------------------------*/
BOOL TestGenericECP( WORD BaseAddr ) {
	BYTE old_ecr, ecr;

	old_ecr = (BYTE)_inp( BaseAddr + ECP_ECR );
	ecr = (BYTE)((old_ecr & 0x1F) | ECP_TEST_MODE);	
	_outp( BaseAddr + ECP_ECR, ecr );
	
	/* check that the mode switch did take */
	if ( (BYTE)_inp( BaseAddr + ECP_ECR ) != ecr ) goto Fail_ECP_Test;
	   
	_outp( BaseAddr + ECP_T_FIFO, 'H' );
	_outp( BaseAddr + ECP_T_FIFO, 'i' );
	if ( (BYTE)_inp( BaseAddr + ECP_T_FIFO ) != 'H' ) goto Fail_ECP_Test;
	if ( (BYTE)_inp( BaseAddr + ECP_T_FIFO ) != 'i' ) goto Fail_ECP_Test;

	/* OK it seems to be an ECP port */
	_outp( BaseAddr + ECP_ECR, old_ecr ); /* restore old ECR */

	return( TRUE );
	
Fail_ECP_Test:
	_outp( BaseAddr + ECP_ECR, old_ecr ); /* restore old ECR */
	return( FALSE );		 
}


/*--------------------------------------------------------------------------*/
/*	FindIntLine																*/
/*--------------------------------------------------------------------------*/
BYTE FindIntLine( WORD BaseAddr ) {
	if ( BaseAddr == 0x378 ) return( 7 );
//	if ( BaseAddr == 0x278 ) return( 5 );
//	if ( BaseAddr == 0x3BC ) return( 5 );
	return( NO_PP_INT_AVAILABLE );
}

/*--------------------------------------------------------------------------*/
/*	FindDMAChannel															*/
/*--------------------------------------------------------------------------*/
BYTE FindDMAChannel( WORD BaseAddr ) {
	if ( BaseAddr == 0x378 ) return( 3 );
//	if ( BaseAddr == 0x278 ) return( 1 );  
//	if ( BaseAddr == 0x3BC ) return( 1 );
	return( NO_DMA_CHANNEL_AVAILABLE );
}



/*--------------------------------------------------------------------------*/
/*	Hook_ISR - Installs the PP Interrupt handler							*/
/*--------------------------------------------------------------------------*/
/* 
	This only works for interrupts 5 and 7, so be very 
	careful (isr = irq + 8) sets the dos interrupt (in 
	protected mode...) enables the pic to 
*/
BOOL Hook_ISR( PPORT *Port ) {
 	WORD picmask;
	int hw_int;
 
	hw_int = (int)Port->IntLine;
 	if ( (hw_int != 5) && (hw_int != 7) ) return( FALSE );

	/* clear down INT data structures					*/
	Port->PP_Callback = NULL;
	Port->DMA_Callback = NULL;
	Port->PP_Context = 0;
	Port->DMA_Context = 0;
	
	if ( hw_int == 5 ) {
		gINT5_Port = Port;
	 	DosSetIntVector( (unsigned)(8 + hw_int), PP_INT5_ISR );
	}
	else {
		gINT7_Port = Port;
		DosSetIntVector( (unsigned)(8 + hw_int), PP_INT7_ISR );
    }
    
 	/* enable the irq */
 	_asm cli ;
 	picmask = (BYTE)_inp( PIC_MASK );
 	picmask &= ~(1<<hw_int);
 	outp( PIC_MASK, picmask );
 	_asm sti ;

	return( TRUE );



}



/*--------------------------------------------------------------------------*/
/*	Unhook_ISR - Uninstalls the PP Interrupt handler	 					*/
/*--------------------------------------------------------------------------*/
BOOL Unhook_ISR( PPORT *Port ) {
	WORD picmask;
	int hw_int;
	void (__interrupt __far *OurVector)();	/* the old ISR vector						*/

	/* clear down INT data structures					*/
	Port->PP_Callback = NULL;
	Port->DMA_Callback = NULL;
	Port->PP_Context = 0;
	Port->DMA_Context = 0;

	hw_int = (int)Port->IntLine;
 	if ( (hw_int != 5) && (hw_int != 7) ) return( FALSE );

	OurVector = DosGetIntVector( (unsigned)(8 + hw_int) );

	if ( Port->OldISRVector ) {
	 	DosSetIntVector( (unsigned)(8 + hw_int), Port->OldISRVector );
	}

	/* disable the irq */
	_asm cli ;
	picmask = (BYTE)_inp( PIC_MASK ) ;
	picmask |= (1<<hw_int);
	outp( PIC_MASK, picmask );
	_asm sti ;

	return( TRUE );
}




/* ====================================================================	*/
/*																		*/
/* Low level port access routines										*/
/*																		*/
/* ====================================================================	*/
void SetDataDirection( PPORT *Port, BYTE NewDirection ) {

	if ( NewDirection == Port->Direction ) return;	/* requested == current */

	if ( NewDirection == FORWARD_DIR ) {
		ClearControlBit( Port, DirBit );		
		Port->Direction = FORWARD_DIR;
	}
	else {
		SetControlBit( Port, DirBit );
		PutData( Port, 0xFF ); 			/* for pseudo bi-dir ports */
		Port->Direction = REVERSE_DIR;
	}
	return;
}

void PutData( PPORT *Port, BYTE Data ) {
	_outp( Port->DataAddr, Data );
}

BYTE GetData( PPORT *Port ) {
	BYTE tmp;
	tmp = (BYTE )(BYTE)_inp( Port->DataAddr );
	return( tmp  );
}

void PutControl( PPORT *Port, BYTE Data ) {
	BYTE Data2;
	Data2 = (BYTE)(Data ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data2 );
    Port->ControlShadow = Data; 
}

BYTE GetControl( PPORT *Port ) {
	return( Port->ControlShadow );
}

void SetControlBit( PPORT *Port, BYTE BitMask ) {
	BYTE Data;
	Data = (BYTE)((Port->ControlShadow |= BitMask) ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data );
}	

BYTE GetControlBit( PPORT *Port, BYTE BitMask ) {
	return( (BYTE )(((BYTE)_inp( Port->ControlAddr )^ CONTROL_INVERSION_MASK) & BitMask) );
}

void ClearControlBit( PPORT *Port, BYTE BitMask ) {
	BYTE Data;
	Data = (BYTE)((Port->ControlShadow &= ~BitMask) ^ CONTROL_INVERSION_MASK);
	_outp( Port->ControlAddr, Data );
}

BYTE GetStatus( PPORT *Port ) {
	return( (BYTE )((BYTE)_inp( Port->StatusAddr )^ STATUS_INVERSION_MASK) );
}

BYTE GetStatusBit( PPORT *Port, BYTE BitMask ) {
	return( (BYTE )(((BYTE)_inp( Port->StatusAddr )^ STATUS_INVERSION_MASK) & BitMask) );
}




__inline BYTE UnscrambleUpload( BYTE Data ) {

	if ( Data & Busy ) {
		return( (BYTE)((Data&0x38)>>3) );
	}
	else {
		return( (BYTE)(((0x40)|(Data&0x38))>>3) );
	   	/* busy (bit 7 ) was not set, so inverted it and return */
	}
}




BOOL LPT1_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	WORD i;
	BYTE __far *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT1, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT1 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );
 
 		WRITECTRL( CONTROL_LPT1, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT1 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}
	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT1, HOSTBUSY_HI );
	return( FALSE );              
}


BOOL LPT2_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	WORD i;
	BYTE __far *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT2, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT2 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );

		WRITECTRL( CONTROL_LPT2, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT2 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}
	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT2, HOSTBUSY_HI );
	return( FALSE );              
}

BOOL LPT3_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes ) {
	WORD i;
	BYTE __far *ptr;
	BYTE tmp_lo, tmp_hi;

	if ( Port->TransferMode != NIBBLE_UPLOAD ) return( FALSE );

	ptr = Data;
	for( i = 0; i < Bytes; i++ ) {
		WRITECTRL( CONTROL_LPT3, HOSTBUSY_LO );
 		WHILE_OUT( (tmp_lo = READSTATUS( STATUS_LPT3 )) & PtrClk );
 		tmp_lo = UnscrambleUpload( tmp_lo );

		WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
 		WHILE_OUT( !(tmp_hi = READSTATUS( STATUS_LPT3 )) & PtrClk );
		*ptr++ = (BYTE )(tmp_lo | (UnscrambleUpload( tmp_hi )<<4) );
	}

	return( TRUE );              

WhileoutError:
	WRITECTRL( CONTROL_LPT3, HOSTBUSY_HI );
	return( FALSE );              
}


