/* Copyright 2000 STMicroelectronics. Free for use under the Gnu Public License */

/*
	Version 0.30
			pport.c


	Project: Vivitar PP Camera

	interface to parallel port, takes care of reading
	writing data and the operation mode of the port
	
	B.Stewart 8/1/96 
	B.Stewart + R. Monteith 5/2/96 
	                                                
	Alterations
	----------- 
	BS	21/02/96	added support for temporal difference data upload
	REM	02/04/96	Rewrite to support full 1284 stuff and nibble mode
*/


#ifndef PP
#define PP


/****************************************************************************/
/*																			*/
/*	Public definitions of PPORT Module								 		*/
/*																			*/
/****************************************************************************/


#define LPT1				0x01
#define LPT2				0x02
#define LPT3				0x03


/* The below represent the hardware capabilities of the parallel port.		*/ 
#ifndef COMPAT_PORT_TYPE
#define NULL_PORT_TYPE		0x00	
#define COMPAT_PORT_TYPE 	0x01		
#define BIDIR_PORT_TYPE		0x02
#define ECP_PORT_TYPE		0x04
#define EPP_PORT_TYPE		0x08
#define DMA_PORT_TYPE		0x10	/* ECP DMA transfers can be used 		*/
#define INT_PORT_TYPE		0x20	/* PP interrupt can be used 			*/
#define FAST_DMA_PORT_TYPE 	0x40	/* Fast DMA transfers can be used 		*/
#endif


/* The below represent the current hardware mode the port is in.			*/
#define COMPAT_HW_MODE		0x01	/* std uni-directional mode				*/
#define BIDIR_HW_MODE		0x02	/* bidirectional mode					*/
#define ECP_HW_MODE			0x03	/* ECP mode 							*/
#define EPP_HW_MODE			0x04	/* 					- NOT YET SUPPORTED	*/
#define ECP_DMA_HW_MODE		0x05	/* ECP DMA mode							*/



/* The below represent the std 1284 data transfer mode currently in use.	*/
#define COMPAT_TRANSFER		0x00	/* 					- NOT YET SUPPORTED	*/
#define NIBBLE_TRANSFER		0x01	/* 1284 mode				 			*/	
#define BIDIR_TRANSFER		0x02  	/* NOT USED 							*/
#define ECP_TRANSFER		0x03	/* 1284 mode				 			*/


/* Special PPC modes.These are invoked by using the 1284 Extensibility Link */
/* Flag during negotiation.													*/
#define UPLOAD_FLAG			0x08
#define NIBBLE_UPLOAD 		(UPLOAD_FLAG|NIBBLE_TRANSFER)	
#define BIDIR_UPLOAD		(UPLOAD_FLAG|BIDIR_TRANSFER)	
#define ECP_UPLOAD			(UPLOAD_FLAG|ECP_TRANSFER)


/* The flag below can be added to UPLOAD modes defined above to specify that*/
/* the difference data should be Run Length Encoded.						*/
#define NORMAL_DATA_FLAG  	0x00	
#define RLE_DATA_FLAG		0x40



#define FORWARD_DIR			0x00
#define REVERSE_DIR			0x01

/* status register */                 
#define S3	0x08   /* pin 15 */
#define S4	0x10   /* pin 13 */
#define S5	0x20   /* pin 12 */
#define	S6	0x40   /* pin 10 */
#define	S7	0x80   /* pin 11   inverted */

/* control register */
#define	C0	0x01   /* pin 1    inverted */
#define	C1	0x02   /* pin 14   inverted */
#define	C2	0x04   /* pin 16 */
#define	C3	0x08   /* pin 17   inverted */
#define C4	0x10   /* IRQ enable, if HI then S6 is connected to IRQ7 */
#define C5	0x20   /* 0 = Fwd, 1 = Reverse */


#define STATUS_INVERSION_MASK	(S7) 		/* for the inversion on some bits */
#define CONTROL_INVERSION_MASK	(C0|C1|C3) 	/* these can be used to correct   */


/* IEEE 1284 Compatiblity Mode signal names 	*/
#define nStrobe		C0	/* inverted */
#define nAutoFd		C1	/* inverted */
#define nInit		C2
#define nSelectIn	C3
#define IntrEnable	C4	/* normally zero for no IRQ */
#define	DirBit		C5	/* 0 = Forward, 1 = Reverse	*/

#define nFault		S3
#define Select		S4
#define PError		S5
#define nAck		S6
#define Busy		S7	/* inverted */	

#pragma warning ( disable : 4121 )

typedef struct {
	/* Data filled in during a PP_DetectPortType 	*/
	BYTE PortType;
	WORD BaseAddr;
	BYTE IntLine;			/* PP interrupt line, normally 7 or 5		*/
	BYTE DMAChannel;		/* ECP DMA channel							*/

 	/* Data filled in by PP_InitialisePPort			*/
	WORD DataAddr;
	WORD StatusAddr;
	WORD ControlAddr;
	WORD ECRAddr;
	WORD DataFIFOAddr;

	BYTE TransferMode;
	BYTE Direction;
	BYTE ControlShadow; 

	WORD Timeout_mS;		/* timeout to which the Whileout counter	*/
							/* will be calibrated to					*/
	DWORD Whileout;

	BOOL EnablePPInt;		/* when this is TRUE the PP interrupt will 	*/
							/* be enabled after a Valid or Immediate .. */
							/* ..1284Termination.						*/

	void (__far *PP_Callback)();
	DWORD PP_Context;
	void (__far *DMA_Callback)();
	DWORD DMA_Context;


	/* Data below is used to restore the port to the state it was 		*/
	/* in before started using it.										*/
	WORD PreviousMode;	   /* which mode ie BIDIR, ECP, EPP				*/
	BYTE PrevData;
	BYTE PrevControl;

	/* state of the ECP interface */
	BYTE PrevECR;
	BYTE PrevCnfgA;
	BYTE PrevCnfgB;
	
	/* old ISR vector													*/
	void (__cdecl __interrupt __far *OldISRVector)();

} PPORT;



/****************************************************************************/
/*																			*/
/*	Private definitions of PPORT Module										*/
/*																			*/
/****************************************************************************/

/* ECP Signal names */
#define HostClk			nStrobe
#define HostAck			nAutoFd
#define nReverseRequest	nInit
#define Active_1284		nSelectIn
#define nPeriphRequest	nFault
#define XFlag			Select
#define nAckReverse		PError
#define PeriphClk		nAck
#define PeriphAck		Busy
#define HOSTACK_LO	( (DirBit | Active_1284 | HostClk)^CONTROL_INVERSION_MASK )
#define HOSTACK_HI	( (DirBit | Active_1284 | HostClk | HostAck)^CONTROL_INVERSION_MASK )


#define LPT1_BASE_ADDR	0x378
#define LPT2_BASE_ADDR	0x278
#define LPT3_BASE_ADDR	0x3BC

/* Offsets from the base address for the various registers */
#define STATUS_OFFSET		0x01
#define CONTROL_OFFSET		0x02

/* define ECP register offsets */
#define ECP_DATA			0x00	/* Data Register			*/
#define ECP_A_FIFO			0x00	/* ECP Address FIFO			*/
#define ECP_DSR				0x01	/* Device Status Register	*/
#define ECP_DCR				0x02	/* Device Control Register	*/
#define ECP_C_FIFO			0x400	/* Compatiblity mode FIFO 	*/
#define ECP_D_FIFO			0x400	/* ECP Data FIFO			*/
#define ECP_T_FIFO			0x400	/* ECP Test FIFO			*/
#define ECP_CNFG_A			0x400	/* Config Register A		*/
#define ECP_CNFG_B			0x401	/* Config Register B		*/
#define ECP_ECR				0x402	/* Extended Control Reg.	*/

#define ECP_FIFO_SIZE	16

/* define some constants for used by the hardcoded functions */
#define BASE_LPT1        LPT1_BASE_ADDR
#define DATA_LPT1		(LPT1_BASE_ADDR)
#define	STATUS_LPT1		(LPT1_BASE_ADDR+STATUS_OFFSET)
#define	CONTROL_LPT1	(LPT1_BASE_ADDR+CONTROL_OFFSET)
#define	DATA_FIFO_LPT1	(LPT1_BASE_ADDR+ECP_D_FIFO)
#define	ECR_LPT1		(LPT1_BASE_ADDR+ECP_ECR)

#define BASE_LPT2        LPT2_BASE_ADDR
#define DATA_LPT2		(LPT2_BASE_ADDR)
#define	STATUS_LPT2		(LPT2_BASE_ADDR+STATUS_OFFSET)
#define	CONTROL_LPT2	(LPT2_BASE_ADDR+CONTROL_OFFSET)
#define	DATA_FIFO_LPT2	(LPT2_BASE_ADDR+ECP_D_FIFO)
#define	ECR_LPT2		(LPT2_BASE_ADDR+ECP_ECR)

#define BASE_LPT3        LPT3_BASE_ADDR
#define DATA_LPT3		(LPT3_BASE_ADDR)
#define	STATUS_LPT3		(LPT3_BASE_ADDR+STATUS_OFFSET)
#define	CONTROL_LPT3	(LPT3_BASE_ADDR+CONTROL_OFFSET)
#define	DATA_FIFO_LPT3	(LPT3_BASE_ADDR+ECP_D_FIFO)
#define	ECR_LPT3		(LPT3_BASE_ADDR+ECP_ECR)


#define WRITECTRL(portaddress,val)		_outp((portaddress),(BYTE )(val))
#define READDATA(portaddress)			(BYTE )_inp((portaddress))                           
#define READSTATUS(portaddress)			(BYTE )_inp((portaddress))                           


/* the following define the bits within the ECR register */
#define ECR_mode			0xE0
#define ECR_nErrIntrEn		0x10
#define ECR_dmaEn			0x08
#define ECR_serviceIntr		0x04
#define ECR_full			0x02
#define ECR_empty			0x01

#define ECP_STANDARD_MODE	0x00
#define ECP_PS2_MODE		0x20
#define ECP_FIFO_MODE		0x40
#define ECP_ECP_MODE		0x60
#define RESERVED_1			0x80
#define RESERVED_2			0xA0
#define ECP_TEST_MODE		0xC0
#define ECP_CNFG_MODE		0xE0




/* Nibble Mode specific names			*/
#define HostClk			nStrobe		
#define HostBusy		nAutoFd

#define nDataAvail		nFault
#define AckDataReq		PError
#define PtrClk			nAck
#define PtrBusy			Busy

#define NIBBLE_DATA_BITS	(nFault|Select|PError|Busy)
#define HOSTBUSY_LO	( (DirBit | Active_1284 | nInit | HostClk)^CONTROL_INVERSION_MASK )
#define HOSTBUSY_HI	( (DirBit | Active_1284 | nInit | HostClk | HostBusy)^CONTROL_INVERSION_MASK )




/* Extensiblity Request Bits 	*/
#define REQUEST_LINK			0x80
#define REQUEST_EPP				0x40
#define REQUEST_ECP_RLE			0x20
#define REQUEST_ECP				0x10
#define REQUEST_DEVICE_ID 		0x04
#define REQUEST_REVERSE_BYTE 	0x01
#define REQUEST_NIBBLE			0x00


/* Valid Extensiblity Request Values */
#define EXTEN_DEVICE_ID			(REQUEST_DEVICE_ID)
#define EXTEN_ECP_MODE 			(REQUEST_ECP)
#define EXTEN_NIBBLE_MODE		(REQUEST_NIBBLE)



#define DEFAULT_TIMEOUT_MS	250	/* default timeout = 250 mS */


/* 8259 programmable interrupt controller (PIC) register addresses  and masks */
#define PIC_CTRL 	0x20             /* PIC control register */
#define PIC_MASK 	0x21             /* PIC mask register   */

#define EOI_CODE  	0x20             /* PIC end of interrupt code */





/****************************************************************************/
/*																			*/
/*	Public functions of PPORT Module																			*/
/*																			*/
/****************************************************************************/

/*
	General Setup Functions
*/
BOOL PP_DetectPortType( BYTE PortNumber, PPORT *Port );


BOOL PP_CreatePPort( PPORT *Port, WORD BaseAddr, BOOL ECPPort, 
								  BYTE IntLine, BYTE DMAChannel );

BOOL PP_InitialisePPort( PPORT *Port );   /* sets port to COMPAT_TRANSFER 		*/
void PP_ClosePPort( PPORT *Port );
BOOL PP_Attempt1284Negotiation( PPORT *Port );

BOOL PP_SendVPP( PPORT *Port, BYTE Bytes, BYTE *Buffer );			

BOOL PP_GetTimeout( PPORT *Port, WORD *Timeout_mS );
BOOL PP_SetTimeout( PPORT *Port, WORD Timeout_mS );

BOOL PP_GetHWInfo( PPORT *Port, WORD *BaseAddr, BYTE *PortType,
				   BYTE *InterruptLine, BYTE *DMAChannel );

BOOL PP_SetHWInfo( PPORT *Port, WORD BaseAddr, BYTE PortType, /* if any of these are zero ignore */
				   BYTE InterruptLine, BYTE DMAChannel );
																


/* The functions below select the standard 1284 transfer modes				*/
BOOL PP_SetTransferMode( PPORT *Port, BYTE TransferMode );
BOOL PP_EndTransferMode( PPORT *Port );

/* The functions below select the special PPC Fast Transfer modes. 			*/
BOOL PP_SetUploadMode( PPORT *Port, BYTE UploadMode, BYTE ModeFlags );
BOOL PP_EndUploadMode( PPORT *Port );

/*
	ECP Mode Public Functions 
*/
BOOL PP_Forward2Reverse( PPORT *Port );
BOOL PP_Reverse2Forward( PPORT *Port );

BOOL PP_ECPWriteBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes );
BOOL PP_ECPReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes );

BOOL PP_SimECPWriteBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes );
BOOL PP_SimECPReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes );

/* 
	1284 Nibble Mode Public Functions 
*/
BOOL PP_NibbleReadBuffer( PPORT *Port, BYTE __far *Data, WORD Bytes );


/* 
	Special PPC Upload Modes - use with PP_Set/EndUploadMode() 	
*/
BOOL PP_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes );
BOOL PP_ECPUpload( PPORT *Port, BYTE __far *Data, WORD Bytes );


/*
	Functions to support ECP Uploads 		
*/
BOOL PP_StartECPUpload( PPORT *Port, BYTE __far *Data );
BOOL PP_FinishECPUpload( PPORT *Port );  

BOOL PP_UnloadFIFO( PPORT *Port, BYTE __far* Data, WORD Bytes );
BOOL PP_EmptyFIFO( PPORT *Port, BYTE __far* Data, WORD *Bytes );

/*
	Functions to support ECP DMA Uploads	
*/
BOOL PP_SetDMABuffer( PPORT *Port, BYTE __far* Data, DWORD PhysicalAddr, WORD BuffSize );
BOOL PP_StartDMAUnloadFIFO( PPORT *Port, WORD Bytes );
BOOL PP_CheckDMAComplete( PPORT *Port );
BOOL PP_AbortDMAUnload( PPORT *Port );


/*	
	Functions to support PP Interrupt - used to signal StreamReady or DMA TC
*/
BOOL PP_SetPPCallback( PPORT *Port, void (__far *PP_Callback)( DWORD ContextParam ), DWORD ContextParam );
BOOL PP_SetDMACallback( PPORT *Port, void (__far *DMA_Callback)( DWORD ContextParam ), DWORD ContextParam );

/* below enables the standard (nAck) interrupt at the end of the 	*/
/* next EndTransferMode() / EndUploadMode() call					*/
BOOL PP_EnablePPInterrupt( PPORT *Port, BOOL EnableInt ); 



/****************************************************************************/
/*																			*/
/*	Private functions of PPORT Module										*/
/*																			*/
/****************************************************************************/


BOOL Hook_ISR( PPORT *Port );
BOOL Unhook_ISR( PPORT *Port );
void __interrupt __far PP_INT5_ISR( void );
void __interrupt __far PP_INT7_ISR( void );


/* 
	Functions to do general IEEE 1284 stuff
*/
void SetCOMPATMode( PPORT *Port );
int Negotiate2SetupPhase( PPORT *Port, BYTE ExtensibilityByte, BYTE LinkMode );
int Valid1284Termination( PPORT *Port );
void Immediate1284Termination( PPORT *Port );

/* 
	ECP Mode
*/
int ECPSetupPhase( PPORT *Port );
int TerminateECPMode( PPORT *Port );

void SetECRMode( PPORT *Port, BYTE Mode );
BYTE GetECRBit( PPORT *Port, BYTE BitMask );
void EmptyFIFO( PPORT *Port );

/*
	Nibble Mode 
*/
int NibbleSetupPhase( PPORT *Port );
int TerminateNibbleMode( PPORT *Port );


/* 
	General functions
*/
BYTE TestPortType( WORD BaseAddr );
BOOL TestGenericECP( WORD BaseAddr );

#define NO_PP_INT_AVAILABLE				69
BYTE FindIntLine( WORD BaseAddr );


#define NO_DMA_CHANNEL_AVAILABLE		69
BYTE FindDMAChannel( WORD BaseAddr );

/*
	Low Level port access routines
*/
void SetDataDirection( PPORT *Port, BYTE Direction ); 

void PutData( PPORT *Port, BYTE Data );
BYTE GetData( PPORT *Port );

void PutControl( PPORT *Port, BYTE Data );
BYTE GetControl( PPORT *Port );

void SetControlBit( PPORT *Port, BYTE BitMask );
BYTE GetControlBit( PPORT *Port, BYTE BitMask );
void ClearControlBit( PPORT *Port, BYTE BitMask );

BYTE GetStatus( PPORT *Port );
BYTE GetStatusBit( PPORT *Port, BYTE BitMask );

/*
	Hard coded upload functions
*/

BYTE UnscrambleNibble( BYTE data );

BOOL LPT1_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes );
BOOL LPT2_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes );
BOOL LPT3_NibbleUpload( PPORT *Port, BYTE __far *Data, WORD Bytes );

#endif


/*
#define WIGGLE_LPT2( x ) _outp(CONTROL_LPT2, (x^_inp(CONTROL_LPT2)))
*/

#define WIGGLE_LPT2( x )
